"""
===================================================
Weights initialization (:mod:`reservoirpy.mat_gen`)
===================================================

Quick tools for weight matrices initialization.

This module provides simple tools for reservoir internal weights
and input/feedback weights initialization. Spectral radius of the
internal weights, input scaling and sparsity are fully parametrizable.

Because most of the architectures developed in Reservoir Computing
involve sparsely-connected neuronal units, the preferred format for all
generated matrices is a :py:mod:`scipy.sparse` format (in most cases *csr*).
Sparse arrays allow fast computations and compact representations of
weights matrices, and remains easily readable. They can be parsed back to
simple Numpy arrays just by calling their ``toarray()`` method.

All functions can take as parameter a :py:class:`numpy.random.Generator`
instance, or a seed number, to ensure reproducibility. Both distribution
of weights and distribution of non-zero connections are controlled with the
seed.

.. autosummary::
   :toctree: generated/

    random_sparse
    uniform
    normal
    bernoulli
    ring
    line
    orthogonal
    cluster
    small_world
    zeros
    ones
    fast_spectral_initialization
    Initializer

Example
=======

Random sparse matrix initializer from uniform distribution,
with spectral radius to 0.9 and connectivity of 0.1.

Matrix creation can be delayed...

.. ipython:: python

    from reservoirpy.mat_gen import random_sparse

    initializer = random_sparse(dist="uniform", sr=0.9, connectivity=0.1)
    matrix = initializer(100, 100)
    print(type(matrix), "\\n", matrix[:5, :5])

...or can be performed right away.

.. ipython:: python

    matrix = random_sparse(100, 100, dist="uniform", sr=0.9, connectivity=0.1)
    print(type(matrix), "\\n", matrix[:5, :5])

Random sparse matrix from Gaussian distribution,
with mean of 0 and variance of 0.5 and an out-degree of 2:

.. ipython:: python

    from reservoirpy.mat_gen import normal

    matrix = normal(7, 10, degree=2, direction="out", loc=0, scale=0.5)
    print(type(matrix), "\\n", matrix)

Dense matrix from Gaussian distribution,
with mean of 0 and variance of 0.5:

.. ipython:: python

    from reservoirpy.mat_gen import normal

    matrix = normal(50, 100, loc=0, scale=0.5)
    print(type(matrix), "\\n", matrix[:5, :5])

Sparse matrix from uniform distribution in [-0.5, 0.5],
with connectivity of 0.9 and input_scaling of 0.3:

.. ipython:: python

    from reservoirpy.mat_gen import uniform

    matrix = uniform(200, 60, low=0.5, high=0.5, connectivity=0.9, input_scaling=0.3)
    print(type(matrix), "\\n", matrix[:5, :5])

Sparse matrix from a Bernoulli random variable
giving 1 with probability p and -1 with probability 1-p,
with p=0.5 (by default) with connectivity of 0.2
and fixed seed, in Numpy format:

.. ipython:: python

    from reservoirpy.mat_gen import bernoulli

    matrix = bernoulli(10, 60, connectivity=0.2, sparsity_type="dense")
    print(type(matrix), "\\n", matrix[:5, :5])


References
==========

    .. [1] C. Gallicchio, A. Micheli, and L. Pedrelli,
           ‘Fast Spectral Radius Initialization for Recurrent
           Neural Networks’, in Recent Advances in Big Data and
           Deep Learning, Cham, 2020, pp. 380–390,
           doi: 10.1007/978-3-030-16841-4_39.
"""

import copy
from functools import partial
from inspect import signature
from typing import Callable, Iterable, Literal, Optional, Union

import numpy as np
from numpy.random import Generator
from scipy import sparse, stats
from scipy.sparse.linalg import ArpackNoConvergence

from .observables import spectral_radius
from .type import global_dtype
from .utils.random import rand_generator

__all__ = [
    "fast_spectral_initialization",
    "random_sparse",
    "uniform",
    "normal",
    "bernoulli",
    "zeros",
    "ones",
]

_epsilon = 1e-8  # used to avoid division by zero when rescaling spectral radius


class Initializer:
    """Base class for initializer functions. Allow updating initializer function
    parameters several times before calling. May perform spectral radius rescaling
    or input scaling as a post-processing to initializer function results.

    Parameters
    ----------
    func : callable
        Initializer function. Should have a `shape` argument and return a Numpy array
        or Scipy sparse array.
    allow_sr : bool, default to True
        Authorize spectral radius rescaling for this initializer.
    allow_input_scaling : bool, default to True
        Authorize input_scaling for this initializer.
    allow_rescaling : bool, default to True
        Authorize any kind of rescaling (spectral radius or input scaling) for this
        initializer.

    Example
    -------
    >>> from reservoirpy.mat_gen import random_sparse
    >>> init_func = random_sparse(dist="uniform")
    >>> init_func = init_func(connectivity=0.1)
    >>> matrix = init_func(5, 5)  # actually creates the matrix
    >>> matrix = random_sparse(5, 5, dist="uniform", connectivity=0.1)  # also creates the matrix
    """

    def __init__(
        self,
        func,
        allow_sr=True,
        allow_input_scaling=True,
        allow_rescaling=True,
    ):
        self._func = func
        self._kwargs = dict()
        self._allow_sr = allow_sr
        self._allow_input_scaling = allow_input_scaling
        self._allow_rescaling = allow_rescaling

        self.__doc__ = func.__doc__
        self.__annotations__ = func.__annotations__
        if self._allow_sr:
            self.__annotations__.update({"sr": float})
        if self._allow_input_scaling:
            self.__annotations__.update({"input_scaling": Union[float, Iterable[float]]})

    def __repr__(self):
        return f"{self._func.__name__}({str(self._kwargs)[1:-1]})"

    def __call__(self, *shape, **kwargs):
        if "sr" in kwargs and not self._allow_sr:
            raise ValueError("Spectral radius rescaling is not supported by this initializer.")

        if "input_scaling" in kwargs and not self._allow_input_scaling:
            raise ValueError("Input scaling is not supported by this initializer.")

        init = copy.deepcopy(self)
        init._kwargs = kwargs | init._kwargs

        if len(shape) > 0:
            if init._allow_rescaling:
                return init._func_post_process(*shape, **init._kwargs)
            else:
                return init._func(*shape, **init._kwargs)
        else:
            if len(kwargs) > 0:
                return init
            else:
                return init._func(**init._kwargs)  # should raise, shape is None

    def _func_post_process(self, *shape, sr=None, input_scaling=None, **kwargs):
        """Post process initializer with spectral radius or input scaling factors."""
        if sr is not None and input_scaling is not None:
            raise ValueError("'sr' and 'input_scaling' parameters are mutually exclusive for a " "given matrix.")

        if sr is not None:
            return _scale_spectral_radius(self._func, shape, sr, **kwargs)
        elif input_scaling is not None:
            return _scale_inputs(self._func, shape, input_scaling, **kwargs)
        else:
            return self._func(*shape, **kwargs)


def _get_rvs(dist: str, random_state: Generator, **kwargs) -> Callable:
    """Get a scipy.stats random variable generator.

    Parameters
    ----------
    dist : str
        A scipy.stats distribution.
    random_state : Generator
        A Numpy random generator.

    Returns
    -------
    scipy.stats.rv_continuous or scipy.stats.rv_discrete
        A scipy.stats random variable generator.
    """
    if dist == "custom_bernoulli":
        return _bernoulli_discrete_rvs(**kwargs, random_state=random_state)
    elif dist in dir(stats):
        distribution = getattr(stats, dist)
        return partial(distribution(**kwargs).rvs, random_state=random_state)
    else:
        raise ValueError(
            f"'{dist}' is not a valid distribution name. " "See 'scipy.stats' for all available distributions."
        )


def _bernoulli_discrete_rvs(p=0.5, value: float = 1.0, random_state: Union[Generator, int] = None) -> Callable:
    """Generator of Bernoulli random variables, equal to +value or -value.

    Parameters
    ----------
    p : float, default to 0.5
        Probability of single success (+value). Single failure (-value) probability
        is (1-p).
    value : float, default to 1.0
        Success value. Failure value is equal to -value.

    Returns
    -------
    callable
        A random variable generator.
    """
    rg = rand_generator(random_state)

    def rvs(size: int = 1):
        return rg.choice([value, -value], p=[p, 1 - p], replace=True, size=size)

    return rvs


def _scale_spectral_radius(w_init, shape, sr, **kwargs):
    """Change the spectral radius of a matrix created with an
    initializer.

    Parameters
    ----------
    w_init : Initializer
        An initializer.
    shape : tuple of int
        Shape of the matrix.
    sr : float
        New spectral radius.
    seed: int or Generator
        A random generator or an integer seed.

    Returns
    -------
    Numpy array or Scipy sparse array
        Rescaled matrix.
    """
    convergence = False

    if "seed" in kwargs:
        seed = kwargs.pop("seed")
    else:
        seed = None
    rng = rand_generator(seed)

    w = w_init(*shape, seed=rng, **kwargs)

    while not convergence:
        # make sure the eigenvalues are reachable.
        # (maybe find a better way to do this on day)
        try:
            current_sr = spectral_radius(w)
            if -_epsilon < current_sr < _epsilon:
                current_sr = _epsilon  # avoid div by zero exceptions.
            w *= sr / current_sr
            convergence = True
        except ArpackNoConvergence:  # pragma: no cover
            # as rng has been used before, a new matrix should be generated
            w = w_init(*shape, seed=rng, **kwargs)

    return w


def _scale_inputs(w_init, shape, input_scaling, **kwargs):
    """Rescale a matrix created with an initializer.

    Parameters
    ----------
    w_init : Initializer
        An initializer.
    shape : tuple of int
        Shape of the matrix.
    input_scaling : float
        Scaling parameter.

    Returns
    -------
    Numpy array or Scipy sparse array
        Rescaled matrix.
    """
    w = w_init(*shape, **kwargs)
    if sparse.issparse(w):
        return w.multiply(input_scaling)
    else:
        return np.multiply(w, input_scaling)


def _random_degree(
    m: int,
    n: int,
    degree: int = 10,
    direction: Literal["in", "out"] = "out",
    format: str = "coo",
    dtype: np.dtype = None,
    random_state: Union[None, int, np.random.Generator, np.random.RandomState] = None,
    data_rvs=None,
):
    """Generate a sparse array of the given shape with randomly distributed values.
    - If `direction=out`, each column has `degree` non-zero values.
    - If `direction=in`, each line has `degree` non-zero values.

    Parameters
    ----------
    m, n : int
        shape of the array
    degree : int, optional
        in-degree or out-degree of each node of the corresponding graph of the
        generated matrix:
    direction : {"in", "out"}, defaults to "out"
        Specify the direction of the `degree` value. Allowed values:
        - "in": `degree` corresponds to in-degrees
        - "out": `degree` corresponds to out-degrees
    dtype : dtype, optional
        type of the returned matrix values.
    random_state : {None, int, `numpy.random.Generator`,
                    `numpy.random.RandomState`}, optional

        If `seed` is None (or `np.random`), the `numpy.random.RandomState`
        singleton is used.
        If `seed` is an int, a new ``RandomState`` instance is used,
        seeded with `seed`.
        If `seed` is already a ``Generator`` or ``RandomState`` instance then
        that instance is used.
        This random state will be used
        for sampling the sparsity structure, but not necessarily for sampling
        the values of the structurally nonzero entries of the matrix.
    data_rvs : callable, optional
        Samples a requested number of random values.
        This function should take a single argument specifying the length
        of the ndarray that it will return. The structurally nonzero entries
        of the sparse random array will be taken from the array sampled
        by this function. By default, uniform [0, 1) random values will be
        sampled using the same random state as is used for sampling
        the sparsity structure.

    Returns
    -------
    res : sparse array

    Notes
    -----
    Only float types are supported for now.

    """
    dtype = np.dtype(dtype)

    if data_rvs is None:  # pragma: no cover
        if np.issubdtype(dtype, np.complexfloating):

            def data_rvs(n):
                return random_state.uniform(size=n) + random_state.uniform(size=n) * 1j

        else:
            data_rvs = partial(random_state.uniform, 0.0, 1.0)
    mn = m * n

    tp = np.intc
    if mn > np.iinfo(tp).max:  # pragma: no cover
        tp = np.int64

    if mn > np.iinfo(tp).max:  # pragma: no cover
        msg = """\
Trying to generate a random sparse array such as the product of dimensions is
greater than %d - this is not supported on this machine
"""
        raise ValueError(msg % np.iinfo(tp).max)

    # each column has `degree` non-zero values
    if direction == "out":
        if not 0 <= degree <= m:
            raise ValueError(f"'degree'={degree} must be between 0 and m={m}.")

        i = np.zeros((n * degree), dtype=tp)
        j = np.zeros((n * degree), dtype=tp)
        for column in range(n):
            ind = random_state.choice(m, size=degree, replace=False)
            i[column * degree : (column + 1) * degree] = ind
            j[column * degree : (column + 1) * degree] = column

    # each line has `degree` non-zero values
    elif direction == "in":
        if not 0 <= degree <= n:
            raise ValueError(f"'degree'={degree} must be between 0 and n={n}.")

        i = np.zeros((m * degree), dtype=tp)
        j = np.zeros((m * degree), dtype=tp)
        for line in range(m):
            ind = random_state.choice(n, size=degree, replace=False)
            i[line * degree : (line + 1) * degree] = line
            j[line * degree : (line + 1) * degree] = ind

    else:
        raise ValueError(f'\'direction\'={direction} must either be "out" or "in".')

    vals = data_rvs(len(i)).astype(dtype, copy=False)
    return sparse.coo_array((vals, (i, j)), shape=(m, n)).asformat(format, copy=False)


def _random_sparse(
    *shape: int,
    dist: str,
    connectivity: float = 1.0,
    dtype: np.dtype = global_dtype,
    sparsity_type: str = "csr",
    seed: Union[int, np.random.Generator] = None,
    degree: Union[int, None] = None,
    direction: Literal["in", "out"] = "out",
    **kwargs,
):
    """Create a random array.

    Parameters
    ----------
    *shape : int, int, ..., optional
        Shape (row, columns, ...) of the matrix.
    dist: str
        A distribution name from :py:mod:`scipy.stats` module, such as "norm" or
        "uniform". Parameters like `loc` and `scale` can be passed to the distribution
        functions as keyword arguments to this function. Usual distributions for
        internal weights are :py:class:`scipy.stats.norm` with parameters `loc` and
        `scale` to obtain weights following the standard normal distribution,
        or :py:class:`scipy.stats.uniform` with parameters `loc=-1` and `scale=2`
        to obtain weights uniformly distributed between -1 and 1.
        Can also have the value "custom_bernoulli". In that case, weights will be drawn
        from a Bernoulli discrete random variable alternating between -1 and 1 and
        drawing 1 with a probability `p` (default `p` parameter to 0.5).
    connectivity: float, default to 1.0
        Also called density of the sparse array. By default, creates dense arrays.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        If connectivity is inferior to 1 and shape is only 2-dimensional, then the
        function will try to use one of the Scipy sparse array format ("csr" or "csc").
        Else, a Numpy array ("dense") will be used.
    seed : optional
        Random generator seed. Default to the global value set with
        :py:func:`reservoirpy.set_seed`.
    degree: int, default to None
        If not None, override the `connectivity` argument and corresponds to the number
        of non-zero values along the axis specified by `direction`
    direction: {"in", "out"}, default to "out"
        If `degree` is not None, specifies the axis along which the `degree` non-zero
        values are distributed.
        - If `direction` is "in", each line will have `degree` non-zero values. In other
        words, each node of the corresponding graph will have `degree` in-degrees
        - If `direction` is "out", each column will have `degree` non-zero values. In
        other words, each node of the corresponding graph will have `degree` out-degrees
    **kwargs : optional
        Arguments for the scipy.stats distribution.

    Returns
    -------
    scipy.sparse array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.
    """

    rng = rand_generator(seed)
    rvs = _get_rvs(dist, **kwargs, random_state=rng)

    if degree is not None:
        if len(shape) != 2:
            raise ValueError(f"Matrix shape must have 2 dimensions, got {len(shape)}: {shape}")
        m, n = shape
        matrix = _random_degree(
            m=m,
            n=n,
            degree=degree,
            direction=direction,
            format=sparsity_type,
            dtype=dtype,
            random_state=rng,
            data_rvs=rvs,
        )
    else:
        if 0 < connectivity > 1.0:
            raise ValueError("'connectivity' must be >0 and <1.")

        if connectivity >= 1.0 or len(shape) != 2:
            matrix = np.zeros(shape, dtype=dtype)
            non_zeros_count = int(connectivity * np.prod(shape))
            non_zero_weights = rvs(size=non_zeros_count).astype(dtype)
            raveled_indices = rng.choice(matrix.size, non_zeros_count, replace=False)
            indices = np.unravel_index(raveled_indices, matrix.shape)
            matrix[indices] = non_zero_weights
        else:
            if "rng" in signature(sparse.random_array).parameters:
                rng_arg = {"rng": rng}
            else:
                # compatibility with scipy<1.15 (& thus Python<3.10). TODO: remove when updating those
                rng_arg = {"random_state": rng.integers(4294967295)}
            matrix = sparse.random_array(
                shape,
                density=connectivity,
                format=sparsity_type,
                **rng_arg,
                data_sampler=rvs,
                dtype=dtype,
            )

    # sparse.random may return np.matrix if format="dense".
    # Only ndarray are supported though, hence the explicit cast.
    if type(matrix) is np.matrix:
        matrix = np.asarray(matrix)

    return matrix


random_sparse = Initializer(_random_sparse)


def _uniform(
    *shape: int,
    low: float = -1.0,
    high: float = 1.0,
    connectivity: float = 1.0,
    dtype: np.dtype = global_dtype,
    sparsity_type: str = "csr",
    seed: Union[int, np.random.Generator] = None,
    degree: Union[int, None] = None,
    direction: Literal["in", "out"] = "out",
):
    """Create an array with uniformly distributed values.

    Parameters
    ----------
    *shape : int, int, ..., optional
        Shape (row, columns, ...) of the array.
    low, high : float, float, default to -1, 1
        Boundaries of the uniform distribution.
    connectivity: float, default to 1.0
        Also called density of the sparse array. By default, creates dense arrays.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        If connectivity is inferior to 1 and shape is only 2-dimensional, then the
        function will try to use one of the Scipy sparse array format ("csr" or "csc").
        Else, a Numpy array ("dense") will be used.
    seed : optional
        Random generator seed. Default to the global value set with
        :py:func:`reservoirpy.set_seed`.
    degree: int, default to None
        If not None, override the `connectivity` argument and corresponds to the number
        of non-zero values along the axis specified by `direction`
    direction: {"in", "out"}, default to "out"
        If `degree` is not None, specifies the axis along which the `degree` non-zero
        values are distributed.
        - If `direction` is "in", each line will have `degree` non-zero values. In other
        words, each node of the corresponding graph will have `degree` in-degrees
        - If `direction` is "out", each column will have `degree` non-zero values. In
        other words, each node of the corresponding graph will have `degree` out-degrees

    Returns
    -------
    Numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.
    """
    if high < low:
        raise ValueError("'high' boundary must be > to 'low' boundary.")
    return _random_sparse(
        *shape,
        dist="uniform",
        loc=low,
        scale=high - low,
        connectivity=connectivity,
        degree=degree,
        direction=direction,
        dtype=dtype,
        sparsity_type=sparsity_type,
        seed=seed,
    )


uniform = Initializer(_uniform)


def _normal(
    *shape: int,
    loc: float = 0.0,
    scale: float = 1.0,
    connectivity: float = 1.0,
    dtype: np.dtype = global_dtype,
    sparsity_type: str = "csr",
    seed: Union[int, np.random.Generator] = None,
    degree: Union[int, None] = None,
    direction: Literal["in", "out"] = "out",
):
    """Create an array with values distributed following a Gaussian distribution.

    Parameters
    ----------
    *shape : int, int, ..., optional
        Shape (row, columns, ...) of the array.
    loc, scale : float, float, default to 0, 1
        Mean and scale of the Gaussian distribution.
    connectivity: float, default to 1.0
        Also called density of the sparse array. By default, creates dense arrays.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        If connectivity is inferior to 1 and shape is only 2-dimensional, then the
        function will try to use one of the Scipy sparse array format ("csr" or "csc").
        Else, a Numpy array ("dense") will be used.
    seed : optional
        Random generator seed. Default to the global value set with
        :py:func:`reservoirpy.set_seed`.
    degree: int, default to None
        If not None, override the `connectivity` argument and corresponds to the number
        of non-zero values along the axis specified by `direction`
    direction: {"in", "out"}, default to "out"
        If `degree` is not None, specifies the axis along which the `degree` non-zero
        values are distributed.
        - If `direction` is "in", each line will have `degree` non-zero values. In other
        words, each node of the corresponding graph will have `degree` in-degrees
        - If `direction` is "out", each column will have `degree` non-zero values. In
        other words, each node of the corresponding graph will have `degree` out-degrees

    Returns
    -------
    Numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.
    """
    return _random_sparse(
        *shape,
        dist="norm",
        loc=loc,
        scale=scale,
        connectivity=connectivity,
        degree=degree,
        direction=direction,
        dtype=dtype,
        sparsity_type=sparsity_type,
        seed=seed,
    )


normal = Initializer(_normal)


def _bernoulli(
    *shape: int,
    p: float = 0.5,
    connectivity: float = 1.0,
    dtype: np.dtype = global_dtype,
    sparsity_type: str = "csr",
    seed: Union[int, np.random.Generator] = None,
    degree: Union[int, None] = None,
    direction: Literal["in", "out"] = "out",
):
    """Create an array with values equal to either 1 or -1. Probability of success
    (to obtain 1) is equal to p.

    Parameters
    ----------
    *shape : int, int, ..., optional
        Shape (row, columns, ...) of the array.
    p : float, default to 0.5
        Probability of success (to obtain 1).
    connectivity: float, default to 1.0
        Also called density of the sparse array. By default, creates dense arrays.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        If connectivity is inferior to 1 and shape is only 2-dimensional, then the
        function will try to use one of the Scipy sparse array format ("csr" or "csc").
        Else, a Numpy array ("dense") will be used.
    seed : optional
        Random generator seed. Default to the global value set with
        :py:func:`reservoirpy.set_seed`.
    degree: int, default to None
        If not None, override the `connectivity` argument and corresponds to the number
        of non-zero values along the axis specified by `direction`
    direction: {"in", "out"}, default to "out"
        If `degree` is not None, specifies the axis along which the `degree` non-zero
        values are distributed.
        - If `direction` is "in", each line will have `degree` non-zero values. In other
        words, each node of the corresponding graph will have `degree` in-degrees
        - If `direction` is "out", each column will have `degree` non-zero values. In
        other words, each node of the corresponding graph will have `degree` out-degrees

    Returns
    -------
    Numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.
    """
    if 1 < p or p < 0:
        raise ValueError("'p' must be <= 1 and >= 0.")
    return _random_sparse(
        *shape,
        p=p,
        dist="custom_bernoulli",
        connectivity=connectivity,
        dtype=dtype,
        sparsity_type=sparsity_type,
        seed=seed,
        degree=degree,
        direction=direction,
    )


bernoulli = Initializer(_bernoulli)


def _ones(*shape: int, dtype: np.dtype = global_dtype, **kwargs):
    """Create an array filled with 1.

    Parameters
    ----------
    *shape : int, int, ..., optional
        Shape (row, columns, ...) of the array.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.

    Returns
    -------
    Numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.
    """
    return np.ones(shape, dtype=dtype)


ones = Initializer(_ones)


def _zeros(*shape: int, dtype: np.dtype = global_dtype, **kwargs):
    """Create an array filled with 0.

    Parameters
    ----------
    *shape : int, int, ..., optional
        Shape (row, columns, ...) of the array.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.

    Returns
    -------
    Numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.

    Note
    ----

    `sr` parameter is not available for this initializer. The spectral radius of a null
    matrix can not be rescaled.
    """
    return np.zeros(shape, dtype=dtype)


zeros = Initializer(_zeros, allow_sr=False)


def _fast_spectral_initialization(
    N: int,
    *args,
    sr: float = None,
    connectivity: float = 1.0,
    dtype: np.dtype = global_dtype,
    sparsity_type: str = "csr",
    seed: Union[int, np.random.Generator] = None,
    degree: Union[int, None] = None,
    direction: Literal["in", "out"] = "out",
):
    """Fast spectral radius (FSI) approach for weights
    initialization [1]_ of square matrices.

    This method is well suited for computation and rescaling of
    very large weights matrices, with a number of neurons typically
    above 500-1000.

    Parameters
    ----------
    N : int, optional
        Shape :math:`N \\times N` of the array.
        This function only builds square matrices.
    connectivity: float, default to 1.0
        Also called density of the sparse array. By default, creates dense arrays.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        If connectivity is inferior to 1 and shape is only 2-dimensional, then the
        function will try to use one of the Scipy sparse array format ("csr" or "csc").
        Else, a Numpy array ("dense") will be used.
    seed : optional
        Random generator seed. Default to the global value set with
        :py:func:`reservoirpy.set_seed`.
    degree: int, default to None
        If not None, override the `connectivity` argument and corresponds to the number
        of non-zero values along the axis specified by `direction`
    direction: {"in", "out"}, default to "out"
        If `degree` is not None, specifies the axis along which the `degree` non-zero
        values are distributed.
        - If `direction` is "in", each line will have `degree` non-zero values. In other
        words, each node of the corresponding graph will have `degree` in-degrees
        - If `direction` is "out", each column will have `degree` non-zero values. In
        other words, each node of the corresponding graph will have `degree` out-degrees

    Returns
    -------
    Numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.

    Note
    ----

    This function was designed for initialization of a reservoir's internal weights.
    In consequence, it can only produce square matrices. If more than one positional
    argument of shape are provided, only the first will be used.

    References
    -----------

    .. [1] C. Gallicchio, A. Micheli, and L. Pedrelli,
           ‘Fast Spectral Radius Initialization for Recurrent
           Neural Networks’, in Recent Advances in Big Data and
           Deep Learning, Cham, 2020, pp. 380–390,
           doi: 10.1007/978-3-030-16841-4_39.
    """
    if connectivity < 0.0 or connectivity > 1.0:
        raise ValueError("'connectivity' must be >0 and <1.")

    if sr is None or connectivity <= 0.0:
        a = 1
    else:
        a = -(6 * sr) / (np.sqrt(12) * np.sqrt((connectivity * N)))

    return _uniform(
        N,
        N,
        low=np.min((a, -a)),
        high=np.max((a, -a)),
        connectivity=connectivity,
        dtype=dtype,
        sparsity_type=sparsity_type,
        seed=seed,
        degree=degree,
        direction=direction,
    )


fast_spectral_initialization = Initializer(
    _fast_spectral_initialization,
    allow_input_scaling=False,
    allow_rescaling=False,
)


def _ring(
    *shape: int,
    weights: Optional[np.ndarray] = None,
    dtype: np.dtype = global_dtype,
    sparsity_type: str = "csr",
    **kwargs,
):
    """Create a lower cyclic shift matrix.

    This is used for ring reservoirs, which have a circular topology
    (each node `n` is connected to the node `n+1 % units`).

    Note that the `connectivity` and `seed` parameters have no effect.

    Parameters
    ----------
    *shape : (int, int), optional
        Shape (row, columns) of the array. Must be a square matrix, i.e. row == columns.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    weights : array of shape (units, ), optional
        If defined, corresponds to the weights of each connection.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        Format of the output matrix. "csr" and "csc" corresponds to the Scipy sparse
        matrix formats, and "dense" corresponds to a regular Numpy array.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    **kwargs : optional
        This argument is kept for compatibility reasons. This is not used.

    Returns
    -------
    Numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.
    """

    if len(shape) != 2 or shape[0] != shape[1]:
        raise ValueError(f"Shape of the ring matrix must be (units, units), got {shape}.")
    units = shape[0]

    if weights is None:
        weights = np.ones((units,), dtype=dtype)
    row = np.roll(np.arange(units, dtype=np.int32), shift=-1)
    col = np.arange(units, dtype=np.int32)

    matrix = sparse.coo_array((weights, (row, col)), shape=(units, units)).asformat(sparsity_type, copy=False)

    if type(matrix) is np.matrix:
        matrix = np.asarray(matrix)

    return matrix


ring = Initializer(_ring)


def _line(
    *shape: int,
    weights: Optional[np.ndarray] = None,
    dtype: np.dtype = global_dtype,
    sparsity_type: str = "csr",
    **kwargs,
):
    """Create a lower shift matrix.

    This is used for line reservoirs, which have a linear topology
    (each node `n` is connected to the node `n+1`, the last node has no successor).

    Note that the `connectivity` and `seed` parameters have no effect.

    Parameters
    ----------
    *shape : (int, int), optional
        Shape (row, columns) of the array. Must be a square matrix, i.e. row == columns.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    weights : array of shape (units-1, ), optional
        If defined, corresponds to the weights of each connection.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        Format of the output matrix. "csr" and "csc" corresponds to the Scipy sparse
        matrix formats, and "dense" corresponds to a regular Numpy array.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    **kwargs : optional
        This argument is kept for compatibility reasons. This is not used.

    Returns
    -------
    Numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.
    """

    if len(shape) != 2 or shape[0] != shape[1]:
        raise ValueError(f"Shape of the ring matrix must be (units, units), got {shape}.")
    units = shape[0]

    if weights is None:
        weights = np.ones((units - 1,), dtype=dtype)
    row = np.arange(1, units, dtype=np.int32)
    col = np.arange(units - 1, dtype=np.int32)

    matrix = sparse.coo_array((weights, (row, col)), shape=(units, units)).asformat(sparsity_type, copy=False)

    if type(matrix) is np.matrix:
        matrix = np.asarray(matrix)

    return matrix


line = Initializer(_line)


def _orthogonal(
    *shape: int,
    dtype: np.dtype = global_dtype,
    seed: Union[int, np.random.Generator] = None,
    **kwargs,
):
    """Create a random orthogonal matrix, drawn from the O(N) Haar distribution (the only uniform distribution on O(N)).

    Note that the `connectivity` parameter in the Reservoir node has no effect with an orthogonal matrix initializer.

    Parameters
    ----------
    *shape : (int, int), optional
        Shape (row, columns) of the array. Must be a square matrix, i.e. row == columns.
    sr : float, optional
        If defined, then will rescale the spectral radius of the matrix to this value.
    input_scaling: float or array, optional
        If defined, then will rescale the matrix using this coefficient or array
        of coefficients.
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    seed : optional
        Random generator seed. Default to the global value set with
        :py:func:`reservoirpy.set_seed`.
    **kwargs : optional
        This argument is kept for compatibility reasons. This is not used.

    Returns
    -------
    numpy array or callable
        If a shape is given to the initializer, then returns a matrix.
        Else, returns a function partially initialized with the given keyword
        parameters, which can be called with a shape and returns a matrix.
    """

    if len(shape) != 2 or shape[0] != shape[1]:
        raise ValueError(f"Shape of the ring matrix must be (units, units), got {shape}.")
    units = shape[0]
    rg = rand_generator(seed)

    matrix = stats.ortho_group.rvs(dim=units, random_state=rg)

    return matrix


orthogonal = Initializer(_orthogonal)


def _cluster(
    *shape: int,
    dtype: np.dtype = global_dtype,
    seed: Union[int, np.random.Generator] = None,
    p_in=0.1,
    p_out=0.01,
    cluster=3,
    distribution="normal",
    sparsity_type="csr",
    **kwargs,
):
    """
    Create a cluster matrix with given distribution and p_in/p_out parameters. Based on the Erdős-Rényi clustered matrix
    algorithm.

    Parameters
    ----------
    *shape : (int, int), optional
        Shape (row, columns) of the array. Must be a square matrix, i.e. row == columns.
    cluster : int, optional
        Number of clusters to create within the matrix. Default is 3.
    p_in : float, optional
        Defines the connectivity within a cluster. Default is 0.1.
    p_out : float, optional
        Defines the connectivity between clusters. Default is 0.01.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        Format of the output matrix. "csr" and "csc" corresponds to the Scipy sparse
        matrix formats, and "dense" corresponds to a regular Numpy array.
    distribution: {"normal", "uniform", "random", "bernoulli"}, default to "normal"
        A distribution name from :py:mod:`scipy.stats` module, such as "norm" or
        "uniform". Parameters like `loc` and `scale` can be passed to the distribution
        functions as keyword arguments to this function. Usual distributions for
        internal weights are :py:class:`scipy.stats.norm` with parameters `loc` and
        `scale` to obtain weights following the standard normal distribution,
        or :py:class:`scipy.stats.uniform` with parameters `loc=-1` and `scale=2`
        to obtain weights uniformly distributed between -1 and 1.
        Can also have the value "custom_bernoulli". In that case, weights will be drawn
        from a Bernoulli discrete random variable alternating between -1 and 1 and
        drawing 1 with a probability `p` (default `p` parameter to 0.5).
    dtype : numpy.dtype, default to numpy.float64
        A Numpy numerical type.
    seed : optional
        Random generator seed. Default to the global value set with
        :py:func:`reservoirpy.set_seed`.
    **kwargs : optional
        This argument is kept for compatibility reasons. This is not used.

    Returns
    -------
    numpy array or callable

    """

    # Check that the shape is divisible by the amount of cluster
    if shape[0] % cluster != 0:
        raise ValueError("Units must be a multiple of the amount of cluster.")

    rng = rand_generator(seed)
    # Define dictionary for corresponding weight distribution matrices
    matrix_dict = dict(normal=normal, uniform=uniform, random=random_sparse, bernoulli=bernoulli)

    # Check for valid distribution
    if distribution not in matrix_dict:
        raise ValueError(
            f"Distribution {distribution} is not supported. Must be 'normal', 'uniform', 'random', 'bernoulli'."
        )

    # Define the global matrix
    matrix = matrix_dict[distribution](
        shape[0],
        shape[1],
        connectivity=p_out,
        dtype=dtype,
        sparsity_type=sparsity_type,
        seed=rng,
    )

    # Define the number of neurons inside each cluster
    n_c = shape[0] // cluster

    # Define the cluster matrix
    c_matrices = matrix_dict[distribution](cluster, n_c, n_c, connectivity=p_in, dtype=dtype, seed=rng)

    # Create the cluster matrix
    for i in range(cluster):
        matrix[i * n_c : i * n_c + n_c, i * n_c : i * n_c + n_c] = c_matrices[i]

    return matrix


cluster = Initializer(_cluster)


def _small_world(
    *shape: int,
    dtype: np.dtype = global_dtype,
    seed: Union[int, np.random.Generator] = None,
    nb_close_neighbours: int = 2,
    proba_rewire: float = 0.1,
    distribution: Literal["normal", "uniform", "random_sparse", "bernoulli"] = "normal",
    sparsity_type: str = "csr",
    **kwargs,
):
    """
    Create a small-world network using the Watts-Strogatz model.
    This function generates a small-world network adjacency matrix with a specified
    number of close neighbours and a probability of rewiring edges.
    The small-world network is characterized by a high clustering coefficient and
    a short average path length, making it suitable for modeling complex networks.
    Small-world networks are often used in various fields such as sociology,
    neuroscience, and computer science to represent systems with local connections.
    See : Watts, D. J.; Strogatz, S. H. (1998). "Collective dynamics of 'small-world' networks"

    Parameters
    ----------
    *shape : (int, int)
        Shape of the matrix. Must be square.
    nb_close_neighbours : int
        Number of close neighbours per node (must be even).
    proba_rewire : float
        Probability of rewiring edges (between 0 and 1).
    dtype : numpy.dtype
        Output matrix dtype.
    seed : int or np.random.Generator
        Random seed or generator.
    distribution : Literal["normal", "uniform", "random_sparse", "bernoulli"], default to "normal"
        Distribution to use for the weights of the connections.
    sparsity_type : {"csr", "csc", "dense"}, default to "csr"
        Format of the output matrix. "csr" and "csc" corresponds to the Scipy sparse
    **kwargs : unused
        For compatibility.

    Returns
    -------
    numpy.ndarray
        Adjacency matrix of the small-world network.
    """

    if len(shape) != 2 or shape[0] != shape[1]:
        raise ValueError(f"Shape of the small-world matrix must be (units, units), got {shape}.")
    if nb_close_neighbours % 2 != 0:
        raise ValueError("nb_close_neighbours must be even.")
    if not (0 <= proba_rewire <= 1):
        raise ValueError("proba_rewire must be between 0 and 1.")

    units = shape[0]
    rng = rand_generator(seed)
    half_neighbours = nb_close_neighbours // 2

    matrix = sparse.lil_array((units, units), dtype=dtype)

    weight_distribution = dict(normal=normal, uniform=uniform, random_sparse=random_sparse, bernoulli=bernoulli)
    if distribution not in weight_distribution:
        raise ValueError(
            f"Distribution {distribution} is not supported. Must be 'normal', 'uniform', 'random_sparse', 'bernoulli'."
        )
    lower_diag_weights = weight_distribution[distribution](
        units * half_neighbours,
        dtype=dtype,
        seed=seed,
        **kwargs,
    )
    upper_diag_weights = weight_distribution[distribution](
        units * half_neighbours,
        dtype=dtype,
        seed=seed,
        **kwargs,
    )

    # Branching nb_close_neighbours nodes to each node (creating the ring lattice)
    indices = np.arange(units)  # (units,)
    offsets = np.arange(1, half_neighbours + 1).reshape(-1, 1)  # (half_neigh, 1)
    i = (indices + offsets) % units  # (units, half_neighbours)
    j = np.tile(indices, (1, half_neighbours))
    i, j = i.flatten(), j.flatten()
    matrix[i, j] = lower_diag_weights
    matrix[j, i] = upper_diag_weights

    # Rewiring edges with probability proba_rewire
    # We only consider the upper triangle of the matrix to avoid double connections
    edges = np.vstack((i, j)).T

    rewire_mask = rng.random(units * half_neighbours) < proba_rewire
    rewired_edges = edges[rewire_mask]

    for i, j in rewired_edges:
        possible_nodes = np.where(np.logical_and(matrix[[i]].toarray().ravel() == 0, np.arange(units) != i))[0]
        if len(possible_nodes) == 0:
            continue
        new_j = rng.choice(possible_nodes)
        matrix[i, new_j] = matrix[i, j]
        matrix[new_j, i] = matrix[j, i]
        matrix[i, j] = 0
        matrix[j, i] = 0
    return matrix.asformat(sparsity_type, copy=False)


small_world = Initializer(_small_world)



# Author: Nathan Trouvain at 16/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from functools import partial
from typing import Callable, Literal, Optional, Sequence, Union

import numpy as np
from numpy.random import Generator

from reservoirpy.utils.data_validation import check_node_input

from ..activationsfunc import get_function
from ..mat_gen import bernoulli, uniform
from ..node import TrainableNode
from ..type import (
    NodeInput,
    State,
    Timeseries,
    Timestep,
    Weights,
    is_array,
    is_multiseries,
)
from ..utils.random import rand_generator


class IPReservoir(TrainableNode):
    """Pool of neurons with random recurrent connexions, tuned using Intrinsic
    Plasticity.

    Intrinsic Plasticity is applied as described in [1]_ and [2]_.

    Reservoir neurons states, gathered in a vector :math:`\\mathbf{x}`, follow
    the update rule below:

    .. math::


        \\mathbf{r}[t+1] = (1 - \\mathrm{lr}) * \\mathbf{r}[t] + \\mathrm{lr}
        * (\\mathbf{W}_{in} \\cdot \\mathbf{u}[t+1]
         + \\mathbf{W} \\cdot \\mathbf{x}[t]
        + \\mathbf{b}_{in})

    .. math::

        \\mathbf{x}[t+1] = f(\\mathbf{a}*\\mathbf{r}[t+1]+\\mathbf{b})

    Parameters :math:`\\mathbf{a}` and :math:`\\mathbf{b}` are updated following two
    different rules:

    - **1.** Neuron activation is tanh:

    In that case, output distribution should be a Gaussian distribution of parameters
    (:math:`\\mu`, :math:`\\sigma`). The learning rule to obtain this output
    distribution is described in [2]_.

    - **2.** Neuron activation is sigmoid:

    In that case, output distribution should be an exponential distribution of
    parameter :math:`\\mu = \\frac{1}{\\lambda}`.
    The learning rule to obtain this output distribution is described in [1]_ and [2]_.

    where:
        - :math:`\\mathbf{x}` is the output activation vector of the reservoir;
        - :math:`\\mathbf{r}` is the internal activation vector of the reservoir;
        - :math:`\\mathbf{u}` is the input timeseries;
        - :math:`f` and :math:`g` are activation functions.

    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    sr : float, optional
        Spectral radius of recurrent weight matrix.
    lr : float or array-like of shape (units,), default to 1.0
        Neurons leak rate. Must be in :math:`[0, 1]`.
    mu : float, default to 0.0
        Mean of the target distribution.
    sigma : float, default to 1.0
        Variance of the target distribution.
    learning_rate : float, default to 5e-4
        Learning rate.
    epochs : int, default to 1
        Number of training iterations.
    input_scaling : float or array-like of shape (features,), default to 1.0.
        Input gain. An array of the same dimension as the inputs can be used to
        set up different input scaling for each feature.
    input_connectivity : float, default to 0.1
        Connectivity of input neurons, i.e. ratio of input neurons connected
        to reservoir neurons. Must be in :math:`]0, 1]`.
    rc_connectivity : float, default to 0.1
        Connectivity of recurrent weight matrix, i.e. ratio of reservoir
        neurons connected to other reservoir neurons, including themselves.
        Must be in :math:`]0, 1]`.
    Win : callable or array-like of shape (units, features), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Input weights matrix or initializer. If a callable (like a function) is
        used,
        then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    W : callable or array-like of shape (units, units), default to :py:func:`~reservoirpy.mat_gen.uniform`
        Recurrent weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units, 1), default to :py:func:`~reservoirpy.mat_gen.bernoulli`
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the
        shape of
        the returned weight matrix.
    activation : {"tanh", "sigmoid"}, default to "tanh"
        Reservoir units activation function.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    seed : int or :py:class:`numpy.random.Generator`, optional
        A random state seed, for noise generation.
    name : str, optional
        Node name.

    References
    ----------
    .. [1] Triesch, J. (2005). A Gradient Rule for the Plasticity of a
           Neuron’s Intrinsic Excitability. In W. Duch, J. Kacprzyk,
           E. Oja, & S. Zadrożny (Eds.), Artificial Neural Networks:
           Biological Inspirations – ICANN 2005 (pp. 65–70).
           Springer. https://doi.org/10.1007/11550822_11

    .. [2] Schrauwen, B., Wardermann, M., Verstraeten, D., Steil, J. J.,
           & Stroobandt, D. (2008). Improving reservoirs using intrinsic
           plasticity. Neurocomputing, 71(7), 1159–1171.
           https://doi.org/10.1016/j.neucom.2007.12.020

    Example
    -------
    >>> from reservoirpy.nodes import IPReservoir
    >>> reservoir = IPReservoir(
    ...                 100, mu=0.0, sigma=0.1, sr=0.95, activation="tanh", epochs=10)

    We can fit the intrinsic plasticity parameters to reach a normal distribution
    of the reservoir activations.
    Using the :py:func:`~reservoirpy.datasets.narma` timeseries:

    >>> from reservoirpy.datasets import narma
    >>> x = narma(1000)
    >>> _ = reservoir.fit(x, warmup=100)
    >>> states = reservoir.run(x)

    .. plot:: ./api/intrinsic_plasticity_example.py

    """

    #: Recurrent weights matrix (:math:`\mathbf{W}`).
    W: Weights
    #: Input weights matrix (:math:`\mathbf{W}_{in}`).
    Win: Weights
    #: Bias vector (:math:`\mathbf{bias}`).
    bias: Weights
    #: Gain of reservoir activation (:math:`\mathbf{a}`).
    a: float
    #: Bias of reservoir activation (:math:`\mathbf{b}`).
    b: float
    #: Leaking rate (1.0 by default) (:math:`\mathrm{lr}`).
    lr: Union[float, np.ndarray]
    #: Spectral radius of W.
    sr: float
    #: Mean of the target distribution (0.0 by default) (:math:`\mu`).
    mu: float
    #: Variance of the target distribution (1.0 by default) (:math:`\sigma`).
    sigma: float
    #: Learning rate (5e-4 by default).
    learning_rate: float
    #: Number of epochs for training (1 by default).
    epochs: int
    #: Input scaling (float or array) (1.0 by default).
    input_scaling: Union[float, Sequence]
    #: Connectivity (or density) of W (0.1 by default).
    rc_connectivity: float
    #: Connectivity (or density) of Win (0.1 by default).
    input_connectivity: float
    #: Activation of the reservoir units (tanh by default) (:math:`f`).
    activation: Literal["tanh", "sigmoid"]
    #: Number of neuronal units in the reservoir.
    units: int
    #: A random state generator.
    rng: Generator

    def __init__(
        self,
        units: Optional[int] = None,
        sr: Optional[float] = None,
        lr: Union[float, np.ndarray] = 1.0,
        mu: float = 0.0,
        sigma: float = 1.0,
        learning_rate: float = 5e-4,
        epochs: int = 1,
        input_scaling: Union[float, Sequence] = 1.0,
        input_connectivity: float = 0.1,
        rc_connectivity: float = 0.1,
        Win: Union[Weights, Callable] = bernoulli,
        W: Union[Weights, Callable] = uniform,
        bias: Union[Weights, Callable] = bernoulli,
        activation: Literal["tanh", "sigmoid"] = "tanh",
        input_dim: Optional[int] = None,
        dtype: type = np.float64,
        seed: Optional[Union[int, np.random.Generator]] = None,
        name=None,
    ):

        self.units = units
        self.sr = sr
        self.lr = lr
        self.mu = mu
        self.sigma = sigma
        self.learning_rate = learning_rate
        self.epochs = epochs
        self.input_scaling = input_scaling
        self.input_connectivity = input_connectivity
        self.rc_connectivity = rc_connectivity
        self.Win = Win
        self.W = W
        self.bias = bias
        self.activation = get_function(activation)

        if activation == "tanh":
            self.gradient = partial(IPReservoir.gaussian_gradients, mu=mu, sigma=sigma)
        elif activation == "sigmoid":
            self.gradient = partial(IPReservoir.exp_gradients, mu=mu)
        else:
            raise ValueError(
                f"Activation '{activation}' must be 'tanh' or 'sigmoid' when " "applying intrinsic plasticity."
            )

        # set input_dim (if possible)
        if input_dim is not None and is_array(Win) and Win.shape[-1] != input_dim:
            raise ValueError(
                f"Both 'input_dim' and 'Win' are set but their dimensions doesn't "
                f"match: {input_dim} != {Win.shape[-1]}."
            )
        self.input_dim = input_dim
        if is_array(Win):
            self.input_dim = np.shape(Win)[-1]

        # set output_dim
        if units is None and not is_array(W):
            raise ValueError("'units' parameter must not be None if 'W' parameter is not " "a matrix.")
        if units is not None and is_array(W) and W.shape[-1] != units:
            raise ValueError(
                f"Both 'units' and 'W' are set but their dimensions doesn't match: " f"{units} != {W.shape[-1]}."
            )
        self.output_dim = units
        if is_array(W):
            self.output_dim = W.shape[-1]
            self.units = W.shape[-1]

        self.dtype = dtype
        self.rng = rand_generator(seed=seed)
        self.name = name

        self.a: np.ndarray
        self.b: np.ndarray
        self.state: State
        self.initialized = False

    def initialize(self, x: Union[NodeInput, Timestep], y: None = None):

        self._set_input_dim(x)

        [Win_rng, W_rng, bias_rng] = self.rng.spawn(3)

        if callable(self.Win):
            self.Win = self.Win(
                self.units,
                self.input_dim,
                input_scaling=self.input_scaling,
                connectivity=self.input_connectivity,
                dtype=self.dtype,
                seed=Win_rng,
            )

        if callable(self.W):
            self.W = self.W(
                self.units,
                self.units,
                sr=self.sr,
                connectivity=self.rc_connectivity,
                dtype=self.dtype,
                seed=W_rng,
            )

        if callable(self.bias):
            self.bias = self.bias(
                self.units,
                connectivity=1.0,
                dtype=self.dtype,
                seed=bias_rng,
            )

        self.a = np.ones((self.output_dim,))
        self.b = np.zeros((self.output_dim,))

        self.state = dict(internal=np.zeros((self.output_dim,)), out=np.zeros((self.output_dim,)))

        self.initialized = True

    def _step(self, state: State, x: Timestep) -> State:
        W = self.W  # NxN
        Win = self.Win  # NxI
        bias = self.bias  # N or float
        f = self.activation
        lr = self.lr
        (internal, external) = state["internal"], state["out"]

        next_state = W @ external + Win @ x + bias
        next_state = (1 - lr) * internal + lr * next_state

        next_external = f(self.a * next_state + self.b)

        return {"internal": next_state, "out": next_external}

    def fit(self, x: NodeInput, y: None = None, warmup: int = 0) -> "IPReservoir":
        check_node_input(x, expected_dim=self.input_dim)

        if not self.initialized:
            self.initialize(x, y)

        for _epoch in range(self.epochs):
            if is_multiseries(x):
                for seq in x:
                    self.partial_fit(seq[warmup:])
            else:
                self.partial_fit(x[warmup:])

        return self

    def partial_fit(self, x: Timeseries):
        for u in x:
            post_state = self.step(u)
            pre_state = self.state["internal"]

            delta_a, delta_b = self.gradient(x=pre_state.T, y=post_state.T, a=self.a)
            self.a += self.learning_rate * delta_a
            self.b += self.learning_rate * delta_b

    def gaussian_gradients(x, y, a, mu, sigma):
        """KL loss gradients of neurons with tanh activation (~ Normal(mu, sigma))."""
        sig2 = sigma**2
        delta_b = -(-(mu / sig2) + (y / sig2) * (2 * sig2 + 1 - y**2 + mu * y))
        delta_a = (1 / a) + delta_b * x
        return delta_a, delta_b

    def exp_gradients(x, y, a, mu):
        """KL loss gradients of neurons with sigmoid activation
        (~ Exponential(lambda=1/mu))."""
        delta_b = 1 - (2 + (1 / mu)) * y + (y**2) / mu
        delta_a = (1 / a) + delta_b * x
        return delta_a, delta_b



# Author: Nathan Trouvain at 19/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from inspect import signature
from typing import Any, Generator, Mapping, Sequence, TypeVar

import numpy as np

from reservoirpy.type import MultiTimeseries, NodeInput, Timeseries, Timestep

from ..node import Node
from ..nodes import Input, Output


def unfold_mapping(
    data_map: dict[Node, MultiTimeseries],
) -> list[dict[Node, Timeseries]]:
    """Convert a mapping of sequence lists into a list of sequence to nodes mappings."""
    # TODO: extensively test this
    seq_numbers = [len(data_map[n]) for n in data_map.keys()]
    if len(np.unique(seq_numbers)) > 1:
        seq_numbers = {n: len(data_map[n]) for n in data_map.keys()}
        raise ValueError(
            f"Found dataset with inconsistent number of sequences for each node. "
            f"Current number of sequences per node: {seq_numbers}"
        )

    # select an input dataset and check
    n_sequences = len(data_map[list(data_map.keys())[0]])

    mapped_sequences: list[dict[str, Timeseries]] = []
    for i in range(n_sequences):
        sequence = {n: data_map[n][i] for n in data_map.keys()}
        mapped_sequences.append(sequence)

    return mapped_sequences


T = TypeVar("T")


def mapping_iterator(*x: Mapping[T, Timeseries]) -> Generator[list[dict[T, Timestep]], Any, None]:
    n_timesteps = x[0][list(x[0].keys())[0]].shape[0]

    for i in range(n_timesteps):
        yield [{k: v[i] for k, v in data.items()} for data in x]


def join_data(*xs: NodeInput) -> NodeInput:
    if isinstance(xs[0], Sequence):
        return [np.concatenate(elements, axis=-1) for elements in zip(*xs)]
    else:
        return np.concatenate(xs, axis=-1)


def data_from_buffer(buffer: np.ndarray, x: NodeInput) -> tuple[np.ndarray, NodeInput]:
    delay = buffer.shape[0]
    if isinstance(x, Sequence):
        buffer_and_x = [np.concatenate((buffer, series), axis=0) for series in x]
        return buffer_and_x[-1][-delay:], [ts[:-delay] for ts in buffer_and_x]
    if x.ndim == 3:
        duplicated_buffer = np.tile(buffer, (x.shape[0], 1, 1))
        buffer_and_x = np.concatenate((duplicated_buffer, x), axis=1)
        return buffer_and_x[-1][-delay:], buffer_and_x[:, :-delay]
    else:
        buffer_and_x = np.concatenate((buffer, x), axis=0)
        return buffer_and_x[-delay:], buffer_and_x[:-delay]


def check_input_output_connections(edges: list[tuple[Node, int, Node]]):
    """Raise a warning if an Input node has an incoming connection or if an
    Output node has an outgoing connection."""
    output_children = [c for p, d, c in edges if d == 0 and isinstance(p, Output)]
    if len(output_children) > 0:
        raise ValueError("An Output node is connected to another Node.")

    input_children = [p for p, d, c in edges if d == 0 and isinstance(c, Input)]
    if len(input_children) > 0:
        raise ValueError("A Node is connected to an Input node.")


def check_unnamed_in_out(model):
    """Raise ValueError if the model has multiple inputs but an input node is not named,
    or if the model has multiple outputs but an output node is not named
    """
    unnamed_inputs = [n for n in model.inputs if n.name is None]
    if len(model.inputs) > 1 and len(unnamed_inputs) > 0:
        raise ValueError(
            f"Model has multiple input nodes but at least one" f" of them is not named: {unnamed_inputs[1:-1]}."
        )
    unnamed_outputs = [n for n in model.outputs if n.name is None]
    if len(model.outputs) > 1 and len(unnamed_outputs) > 0:
        raise ValueError(
            f"Model has multiple output nodes but at least one" f" of them is not named: {unnamed_outputs[1:-1]}."
        )


def check_unnamed_trainable(model):
    """Raise ValueError if the model has multiple trainable nodes but one of
    them is not named.
    """
    unnamed_nodes = [n for n in model.trainable_nodes if n.name is None]
    if len(model.trainable_nodes) > 1 and len(unnamed_nodes) > 0:
        raise ValueError(
            f"Model has multiple trainable nodes but at least one" f" of them is not named: {unnamed_nodes[1:-1]}."
        )


def obj_from_kwargs(klas, kwargs):
    sig = signature(klas.__init__)
    params = list(sig.parameters.keys())
    klas_kwargs = {n: v for n, v in kwargs.items() if n in params}
    return klas(**klas_kwargs)



# Author: Nathan Trouvain at 12/07/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from typing import Optional, Sequence, Union

from ..node import Node
from ..type import NodeInput, State, Timeseries, Timestep


class Input(Node):
    """Node feeding input data to other nodes in the models.

    Allow creating an input source and connecting it to several nodes at once.

    This node has no parameters and no hyperparameters.

    Parameters
    ----------
    name : str, optional
        Node name.

    Example
    -------

    An input source feeding three different nodes in parallel.

    >>> from reservoirpy.nodes import Reservoir, Input
    >>> source = Input()
    >>> res1, res2, res3 = Reservoir(100), Reservoir(100), Reservoir(100)
    >>> model = source >> [res1, res2, res3]

    A model with different input sources. Use names to identify each source at runtime.

    >>> import numpy as np
    >>> from reservoirpy.nodes import Reservoir, Input
    >>> source1, source2 = Input(name="s1"), Input(name="s2")
    >>> res1, res2 = Reservoir(100, name="res1"), Reservoir(100, name="res2)
    >>> model = source1 >> [res1, res2] & source2 >> [res1, res2]
    >>> outputs = model.run({"s1": np.ones((10, 5)), "s2": np.ones((10, 3))})
    """

    initialized: bool
    state: State

    def __init__(self, name: Optional[str] = None):
        self.initialized = False
        self.name = name
        self.state = {}

    def initialize(self, x: Union[NodeInput, Timestep]):
        self._set_input_dim(x)
        self.output_dim = self.input_dim
        self.initialized = True

    def _step(self, state: State, x: Timestep) -> State:
        return {"out": x}

    def _run(self, state: State, x: Timeseries) -> tuple[State, Timeseries]:
        return {"out": x[-1]}, x


class Output(Node):
    """Convenience node which can be used to add an output to a model.

    For instance, this node can be connected to a reservoir within a model to inspect
    its states.

    Parameters
    ----------
    name : str, optional
        Node name.

    Example
    -------

    We can use the :py:class:`Output` node to probe the hidden states of Reservoir
    in an Echo State Network:

    >>> import numpy as np
    >>> from reservoirpy.nodes import Reservoir, Ridge, Output
    >>> reservoir = Reservoir(100)
    >>> readout = Ridge(name="readout")
    >>> probe = Output(name="reservoir-states")
    >>> esn = reservoir >> readout & reservoir >> probe
    >>> _ = esn.initialize(np.ones((1,1)), np.ones((1,1)))

    When running the model, states can then be retrieved as an output:

    >>> data = np.ones((10, 1))
    >>> outputs = esn.run(data)
    >>> states = outputs["reservoir-states"]
    """

    initialized: bool
    state: State

    def __init__(self, name: Optional[str] = None):
        self.initialized = False
        self.name = name
        self.state = {}

    def initialize(self, x: Union[NodeInput, Timestep]):
        dim = x.shape[-1] if not isinstance(x, Sequence) else x[0].shape[-1]
        self.input_dim = dim
        self.output_dim = dim
        self.initialized = True

    def _step(self, state: State, x: Timestep) -> State:
        return {"out": x}

    def _run(self, state: State, x: Timeseries) -> tuple[State, Timeseries]:
        return {"out": x[-1]}, x



# Author: Nathan Trouvain at 10/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest

from .dummy_nodes import MinusNode, Offline, PlusNode


def test_node_link():
    plus_node = PlusNode()
    minus_node = MinusNode()
    offline1 = Offline()
    offline2 = Offline()

    model1 = plus_node >> minus_node
    model2 = minus_node >> plus_node

    assert model1.edges == [(plus_node, 0, minus_node)]
    assert model2.edges == [(minus_node, 0, plus_node)]
    assert set(model1.nodes) == set(model2.nodes)

    model3 = plus_node >> offline1
    model4 = minus_node >> offline2

    model = model3 >> model4

    assert model.edges == [
        (plus_node, 0, offline1),
        (minus_node, 0, offline2),
        (offline1, 0, minus_node),
    ]
    assert set(model.nodes) == set(model3.nodes) | set(model4.nodes)

    # cycles in the model!
    with pytest.raises(RuntimeError):
        _ = model1 & model2

    with pytest.raises(RuntimeError):
        _ = plus_node >> minus_node >> plus_node

    with pytest.raises(RuntimeError):
        _ = plus_node >> plus_node

    x = np.ones((5,))
    x2 = np.ones((6,))
    plus_node(x)
    minus_node(x2)

    # bad dimensions
    with pytest.raises(ValueError):
        _ = plus_node >> minus_node

    with pytest.raises(ValueError):
        _ = model1(x)

    # merge inplace on a node
    model = plus_node
    model &= minus_node
    assert model.nodes == [plus_node, minus_node]
    assert model.edges == []


def test_node_link_several():
    plus_node = PlusNode(name="Plus")
    minus_node = MinusNode(name="Minus")
    offline_node = Offline(name="Offline")
    model = [plus_node, minus_node] >> offline_node

    assert len(model.nodes) == 3
    assert len(model.edges) == 2

    model = plus_node >> [offline_node, minus_node]

    assert model.nodes == [plus_node, offline_node, minus_node]
    assert model.edges == [(plus_node, 0, offline_node), (plus_node, 0, minus_node)]


def test_model_merge():
    plus_node = PlusNode(name="Plus")
    minus_node = MinusNode(name="Minus")
    offline_node = Offline(name="Offline")

    branch1 = plus_node >> minus_node
    branch2 = plus_node >> offline_node

    model = branch1 & branch2

    assert set(model.nodes) == {plus_node, minus_node, offline_node}
    assert set(model.edges) == {
        (plus_node, 0, minus_node),
        (plus_node, 0, offline_node),
    }

    branch1 &= branch2

    assert set(branch1.nodes) == {plus_node, minus_node, offline_node}
    assert set(branch1.edges) == {
        (plus_node, 0, minus_node),
        (plus_node, 0, offline_node),
    }



# Author: Nathan Trouvain at 12/07/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from collections import deque
from typing import Any, Optional, Sequence, TypeVar

T = TypeVar("T")


def unique_ordered(x: Sequence[T]) -> list[T]:
    """Returns a list with the same elements as x occuring only once, and in the same order"""
    return list(dict.fromkeys(x))


def find_parents_and_children(nodes: list[T], edges: list[tuple[T, int, T]]):
    """Returns two dicts linking nodes to their parents and children in the graph."""
    parents = {
        child: unique_ordered([p for p, d, c in edges if c is child and d == 0])
        for child in nodes
    }
    children = {
        parent: unique_ordered([c for p, d, c in edges if p is parent and d == 0])
        for parent in nodes
    }

    return parents, children


def find_indirect_children(nodes: list[T], edges: list[tuple[T, int, T]]):
    """Returns two dicts linking nodes to their children in the graph."""

    children = {
        parent: unique_ordered([c for p, d, c in edges if p is parent])
        for parent in nodes
    }

    return children


def topological_sort(
    nodes: list[T], edges: list[tuple[T, int, T]], inputs: Optional[list[T]] = None
) -> list[T]:
    """Topological sort of nodes in a Model, to determine execution order."""
    if inputs is None:
        inputs = find_inputs(nodes, edges)

    parents, children = find_parents_and_children(nodes, edges)

    # using Kahn's algorithm
    ordered_nodes = []
    edges_set = set(edges)
    inputs_deque = deque(inputs)
    while len(inputs_deque) > 0:
        n = inputs_deque.pop()
        ordered_nodes.append(n)
        for m in children.get(n, ()):
            edges_set.remove((n, 0, m))
            parents[m].remove(n)
            if parents.get(m) is None or len(parents[m]) < 1:
                inputs_deque.append(m)
    if len(edges_set) > 0:
        raise RuntimeError(
            "Model has a cycle: impossible "
            "to automatically determine operations "
            "order in the model."
        )
    else:
        return ordered_nodes


def find_pseudo_inputs(
    nodes: list[T], edges: list[tuple[T, int, T]], y_mapping: dict[T, Any]
) -> list[T]:
    """
    Find all nodes that are not receivers or that only receive forced feedback.
    Guaranteed to preserve order.
    """
    unsupervised_receivers: set[T] = set(
        [n for c, d, n in edges if d == 0 and c not in y_mapping.keys()]
    )
    sources = [node for node in nodes if node not in unsupervised_receivers]
    return sources


def find_inputs(nodes: list[T], edges: list[tuple[T, int, T]]) -> list[T]:
    """
    Find all nodes that are not receivers (without direct incoming connections).
    Guaranteed to preserve order.
    """
    receivers: set[T] = set([n for _, d, n in edges if d == 0])
    sources = [node for node in nodes if node not in receivers]
    return sources


def find_outputs(nodes: list[T], edges: list[tuple[T, int, T]]) -> list[T]:
    """
    Find all nodes that are not senders (no direct out-going connections).
    Guaranteed to preserve order.
    """
    senders = set([n for n, d, _ in edges if d == 0])
    sinks = [node for node in nodes if node not in senders]
    return sinks



# Author: Nathan Trouvain at 17/05/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal

from reservoirpy.nodes import LMS, Reservoir


def test_lms_init():
    X = np.ones((2, 100))
    Y = np.ones((2, 10))

    # with callable Wout & bias
    node = LMS(0.1)
    assert node.input_dim is None
    assert node.output_dim is None
    node.initialize(X, Y)
    assert_array_equal(node.Wout, np.zeros((100, 10)))
    assert_array_equal(node.bias, np.zeros((10,)))
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.learning_rate == 0.1

    # with initialized weights
    node = LMS(1e-6, Wout=np.ones((100, 10)), bias=np.ones((10,)))
    node.initialize(X, Y)
    assert_array_equal(node.Wout, np.ones((100, 10)))
    assert_array_equal(node.bias, np.ones((10,)))
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.learning_rate == 1e-6


def test_lms_fit():
    node = LMS(1e-3)

    x = np.ones((7, 5, 2))
    y = np.ones((7, 5, 10))

    res = node.fit(x, y)

    assert node.Wout.shape == (2, 10)
    assert not np.all(node.Wout == np.zeros((2, 10)))
    assert node.bias.shape == (10,)
    assert not np.all(node.bias == np.zeros((10,)))
    assert node.learning_rate == 1e-3

    data = np.ones((1000, 2))
    res = node.run(data)

    assert res.shape == (1000, 10)


def test_lms_partial_fit():
    rng = np.random.default_rng(seed=2368)
    X = rng.normal(size=(10, 5, 2))
    Y = rng.normal(size=(10, 5, 10))
    data = rng.normal(size=(1000, 2))

    node1 = LMS(1e-4)
    for x, y in zip(X, Y):
        node1.partial_fit(x, y)

    assert node1.Wout.shape == (2, 10)
    assert node1.bias.shape == (10,)

    res1 = node1.run(data)
    assert res1.shape == (1000, 10)

    # compare with .fit()
    node2 = LMS(1e-4)
    node2.fit(X.reshape(10 * 5, 2), Y.reshape(10 * 5, 10))
    res2 = node2.run(data)
    assert_array_almost_equal(res1, res2)
    assert not np.all(res1 == np.zeros((1000, 10)))
    assert_array_almost_equal(node1.Wout, node2.Wout)
    assert not np.all(node1.Wout == np.zeros((2, 10)))
    assert_array_almost_equal(node1.bias, node2.bias)
    assert not np.all(node1.bias == np.zeros((10,)))


def test_lms_call():
    x = np.array([0.1, 0.2, 0.3])

    node = LMS(output_dim=3)
    y1 = node.step(x)
    assert_array_equal(y1, np.zeros((3,)))
    node = LMS(output_dim=2)
    y2 = node(x)
    assert_array_equal(y2, np.zeros((2,)))



"""
======================================
Datasets (:mod:`reservoirpy.datasets`)
======================================

.. currentmodule:: reservoirpy.datasets

Chaotic timeseries on continuous time
=====================================

Timeseries expressing a chaotic behaviour and defined on a continuous
time axis, generated at will.

All timeseries defined by differential equations on a
continuous space are by default approximated using 4-5th order
Runge-Kuta method [1]_, either homemade (for Mackey-Glass timeseries)
or from Scipy :py:func:`scipy.integrate.solve_ivp` tool.

.. autosummary::
   :toctree: generated/

    mackey_glass - Mackey-Glass delayed differential equations timeseries.
    lorenz - Lorenz system timeseries.
    multiscroll - Multi scroll attractor timeseries.
    doublescroll - Double scroll attractor timeseries.
    rabinovich_fabrikant - Rabinovitch-Fabrikant differential equations timeseries.
    lorenz96 - Lorenz 1996 attractor.
    rossler - Rossler attractor.
    kuramoto_sivashinsky - Kuramoto-Sivashinsky oscillators.
    mso - Multiple superimposed oscillators.
    mso2 - Multiple superimposed oscillators with 2 frequencies.
    mso8 - Multiple superimposed oscillators with 8 frequencies.

Chaotic timeseries on discrete time
===================================

Timeseries expressing a chaotic behaviour and defined on a discrete
time axis, generated at will.

Discrete timeseries are defined using recurrent time-delay relations.

.. autosummary::
    :toctree: generated/

    logistic_map - Logistic map timeseries
    henon_map - Hénon map timeseries
    narma - NARMA timeseries


Classification/pattern recognition tasks
========================================

Classified datasets of timeseries.

.. autosummary::
    :toctree: generated/

    japanese_vowels - Japanese vowels task


Miscellaneous
=============

.. autosummary::
    :toctree: generated/

    santafe_laser - Santa-Fe laser dataset
    to_forecasting - Timeseries splitting utility
    one_hot_encode - One-hot encoding utility
    from_aeon_classification - Conversion from Aeon's format
    set_seed - Change random seed for dataset generation
    get_seed - Return random seed used for dataset generation


References
==========

    .. [1] `Runge–Kutta methods
           <https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods>`_
           on Wikipedia.
    .. [2] M. Hénon, ‘A two-dimensional mapping with a strange
           attractor’, Comm. Math. Phys., vol. 50, no. 1, pp. 69–77, 1976.
    .. [3] `Hénon map <https://en.wikipedia.org/wiki/H%C3%A9non_map>`_
           on Wikipedia
    .. [4] R. M. May, ‘Simple mathematical models with very
           complicated dynamics’, Nature, vol. 261, no. 5560,
           Art. no. 5560, Jun. 1976, doi: 10.1038/261459a0.
    .. [5] `Logistic map <https://en.wikipedia.org/wiki/Logistic_map>`_
           on Wikipedia
    .. [6] E. N. Lorenz, ‘Deterministic Nonperiodic Flow’,
           Journal of the Atmospheric Sciences, vol. 20, no. 2,
           pp. 130–141, Mar. 1963,
           doi: 10.1175/1520-0469(1963)020<0130:DNF>2.0.CO;
    .. [7] `Lorenz system <https://en.wikipedia.org/wiki/Lorenz_system>`_
           on Wikipedia.
    .. [8] M. C. Mackey and L. Glass, ‘Oscillation and chaos in
           physiological control systems’, Science, vol. 197, no. 4300,
           pp. 287–289, Jul. 1977, doi: 10.1126/science.267326.
    .. [9] `Mackey-Glass equations
            <https://en.wikipedia.org/wiki/Mackey-Glass_equations>`_
            on Wikipedia.
    .. [10] G. Chen and T. Ueta, ‘Yet another chaotic attractor’,
            Int. J. Bifurcation Chaos, vol. 09, no. 07, pp. 1465–1466,
            Jul. 1999, doi: 10.1142/S0218127499001024.
    .. [11] `Chen double scroll attractor
            <https://en.wikipedia.org/wiki/Multiscroll_attractor
            #Chen_attractor>`_ on Wikipedia.
    .. [12] M. I. Rabinovich and A. L. Fabrikant,
            ‘Stochastic self-modulation of waves in
            nonequilibrium media’, p. 8, 1979.
    .. [13] `Rabinovich-Fabrikant equations
            <https://en.wikipedia.org/wiki/Rabinovich%E2%80
            %93Fabrikant_equations>`_
            on Wikipedia.
    .. [14] A. F. Atiya and A. G. Parlos, ‘New results on recurrent
            network training: unifying the algorithms and accelerating
            convergence,‘ in IEEE Transactions on Neural Networks,
            vol. 11, no. 3, pp. 697-709, May 2000,
            doi: 10.1109/72.846741.
    .. [15] B.Schrauwen, M. Wardermann, D. Verstraeten, J. Steil,
            D. Stroobandt, ‘Improving reservoirs using intrinsic
            plasticity‘,
            Neurocomputing, 71. 1159-1171, 2008,
            doi: 10.1016/j.neucom.2007.12.020.
    .. [16] M. Kudo, J. Toyama and M. Shimbo. (1999).
            "Multidimensional Curve Classification Using Passing-Through Regions".
            Pattern Recognition Letters, Vol. 20, No. 11--13, pages 1103--1111.
    .. [17] Lorenz, E. N. (1996, September).
            Predictability: A problem partly solved. In Proc.
            Seminar on predictability (Vol. 1, No. 1).
    .. [18] O.E. Rössler, "An equation for continuous chaos", Physics Letters A,
            vol 57, Issue 5, Pages 397-398, ISSN 0375-9601, 1976,
            https://doi.org/10.1016/0375-9601(76)90101-8.
    .. [19] Kuramoto, Y. (1978). Diffusion-Induced Chaos in Reaction Systems.
            Progress of Theoretical Physics Supplement, 64, 346–367.
            https://doi.org/10.1143/PTPS.64.346
    .. [20] Sivashinsky, G. I. (1977). Nonlinear analysis of hydrodynamic
            instability in laminar flames—I. Derivation of basic equations.
            Acta Astronautica, 4(11), 1177–1206.
            https://doi.org/10.1016/0094-5765(77)90096-0
    .. [21] Sivashinsky, G. I. (1980). On Flame Propagation Under Conditions
            of Stoichiometry. SIAM Journal on Applied Mathematics, 39(1), 67–82.
            https://doi.org/10.1137/0139007
    .. [22] Jaeger, H. (2004b). Seminar slides. (Online) Available
            http://www.faculty.iu-bremen.de/hjaeger/courses/SeminarSpring04/ESNStandardSlides.pdf.
    .. [24] Roeschies, B., & Igel, C. (2010). Structure optimization of
            reservoir networks. Logic Journal of IGPL, 18(5), 635-669.
    .. [25] Weigend, A. S. (2018). Time series prediction: forecasting the
            future and understanding the past. Routledge.
"""

from typing import Optional, Union

import numpy as np

from ._chaos import (
    doublescroll,
    henon_map,
    kuramoto_sivashinsky,
    logistic_map,
    lorenz,
    lorenz96,
    mackey_glass,
    multiscroll,
    narma,
    rabinovich_fabrikant,
    rossler,
)
from ._japanese_vowels import japanese_vowels
from ._santafe_laser import santafe_laser
from ._seed import get_seed, set_seed
from ._utils import from_aeon_classification, one_hot_encode

__all__ = [
    "henon_map",
    "logistic_map",
    "lorenz",
    "mackey_glass",
    "multiscroll",
    "rabinovich_fabrikant",
    "narma",
    "doublescroll",
    "japanese_vowels",
    "lorenz96",
    "rossler",
    "kuramoto_sivashinsky",
    "set_seed",
    "get_seed",
    "to_forecasting",
    "mso",
    "mso2",
    "mso8",
    "from_aeon_classification",
    "one_hot_encode",
    "santafe_laser",
]


def to_forecasting(
    timeseries: np.ndarray,
    forecast: int = 1,
    axis: int = 0,
    test_size: Optional[Union[int, float]] = None,
):
    """Split a timeseries for forecasting tasks.

    Transform a timeseries :math:`X` into a series of
    input values :math:`X_t` and a series of output values
    :math:`X_{t+\\mathrm{forecast}}`.

    It is also possible to split the timeseries between training
    timesteps and testing timesteps.

    Parameters
    ----------
    timeseries : np.ndarray
        Timeseries to split.
    forecast : int, optional
        Number of time lag steps between
        the timeseries :math:`X_t` and the timeseries
        :math:`X_{t+\\mathrm{forecast}}`, by default 1,
        i.e. returns two timeseries with a time difference
        of 1 timesteps.
    axis : int, optional
        Time axis of the timeseries, by default 0
    test_size : int or float, optional
        If set, will also split the timeseries
        into a training phase and a testing phase of
        ``test_size`` timesteps. Can also be specified
        as a float ratio, by default None

    Returns
    -------
    tuple of numpy.ndarray
        :math:`X_t` and :math:`X_{t+\\mathrm{forecast}}`.

        If ``test_size`` is specified, will return:
        :math:`X_t`, :math:`X_t^{test}`,
        :math:`X_{t+\\mathrm{forecast}}`, :math:`X_{t+\\mathrm{forecast}}^{test}`.

        The size of the returned timeseries is therefore the size of
        :math:`X` minus the forecasting length ``forecast``.

    Examples
    --------

    >>> from reservoirpy.datasets import to_forecasting
    >>> from numpy import linspace, sin
    >>> t = linspace(0, 500, 1000)
    >>> timeseries = sin(0.2*t) + sin(0.311*t)
    >>> X_train, X_test, y_train, y_test = to_forecasting(timeseries, forecast=10, test_size=200)

    .. plot::

        from reservoirpy.datasets import to_forecasting
        from numpy import linspace, sin
        t = linspace(0, 500, 1000); h=10
        timeseries = sin(0.2*t) + sin(0.311*t)
        X_train, X_test, y_train, y_test = to_forecasting(timeseries, forecast=h, test_size=200)
        plt.figure(figsize=(12, 3))
        plt.plot(t[:-200-h], X_train, label="Training data")
        plt.plot(t[:-200-h], y_train, label="Training ground truth")
        plt.plot(t[-200:], X_test, label="Testing data")
        plt.plot(t[-200:], y_test, label="Testing ground truth")
        plt.legend(loc="lower left", bbox_to_anchor=(0.0, 0.0))
        plt.show()

    Raises
    ------
    ValueError
        If ``test_size`` is a float, it must be in [0, 1[.
    """

    series_ = np.moveaxis(timeseries.view(), axis, 0)
    time_len = series_.shape[0]

    if test_size is not None:
        if isinstance(test_size, float) and test_size < 1 and test_size >= 0:
            test_len = round(time_len * test_size)
        elif isinstance(test_size, int):
            test_len = test_size
        else:
            raise ValueError(
                "invalid test_size argument: "
                "test_size can be an integer or a float "
                f"in [0, 1[, but is {test_size}."
            )
    else:
        test_len = 0

    X = series_[:-forecast]
    y = series_[forecast:]

    if test_len > 0:
        X_t = X[-test_len:]
        y_t = y[-test_len:]
        X = X[:-test_len]
        y = y[:-test_len]

        X = np.moveaxis(X, 0, axis)
        X_t = np.moveaxis(X_t, 0, axis)
        y = np.moveaxis(y, 0, axis)
        y_t = np.moveaxis(y_t, 0, axis)

        return X, X_t, y, y_t

    return np.moveaxis(X, 0, axis), np.moveaxis(y, 0, axis)


def mso(n_timesteps: int, freqs: list, normalize: bool = True):
    """Multiple superimposed oscillator task [22]_

    This task is usually performed to evaluate a model resistance
    to perturbations and its asymptotic stability. See, for example: [23]_.

    .. math::

        MSO(t) = \\sum_{i} sin(f_i t)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate
    freqs : list
        Frequencies of the sin waves
    normalize : bool, optional
        If `True`, scales the range of values in [-1, 1]

    Returns
    -------
    array of shape (n_timesteps, 1)
        Multiple superimposed oscillator timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import mso
    >>> timeseries = mso(200, freqs=[0.2, 0.44])
    >>> print(timeseries.shape)
    (200, 1)

    .. plot::

        from reservoirpy.datasets import mso
        timeseries = mso(200, freqs=[0.2, 0.44])
        plt.figure(figsize=(12, 4))
        plt.plot(timeseries)
        plt.show()

    References
    ----------
    .. [22] Jaeger, H. (2004b). Seminar slides. (Online) Available http://www.faculty.
            iu-bremen.de/hjaeger/courses/SeminarSpring04/ESNStandardSlides.pdf.
    .. [23] Jaeger, H., Lukoševičius, M., Popovici, D., & Siewert, U. (2007).
            Optimization and applications of echo state networks with leaky-integrator
            neurons. Neural networks, 20(3), 335-352.
    """
    t = np.arange(n_timesteps).reshape(n_timesteps, 1)
    y = np.zeros((n_timesteps, 1))

    for f in freqs:
        y += np.sin(f * t)

    if normalize:
        return (2 * y - y.min() - y.max()) / (y.max() - y.min())
    else:
        return y


def mso2(n_timesteps: int, normalize: bool = True):
    """Multiple superimposed oscillator task with 2 frequencies [22]_

    This is the $MSO$ task with 2 frequencies: :math:`f_1=0.2` and
    :math:`f_1=0.311`, as first described in [22]_.

    .. math::

        MSO(t) = sin(0.2 t) + sin(0.311 t)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate
    normalize : bool
        If `True`, scales the range of values in [-1, 1]

    Returns
    -------
    array of shape (n_timesteps, 1)
        Multiple superimposed oscillator timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import mso2
    >>> timeseries = mso(500)
    >>> print(timeseries.shape)
    (500, 1)

    .. plot::

        from reservoirpy.datasets import mso2
        timeseries = mso2(500)
        plt.figure(figsize=(12, 4))
        plt.plot(timeseries)
        plt.show()

    References
    ----------
    .. [22] Jaeger, H. (2004b). Seminar slides. (Online) Available http://www.faculty.
            iu-bremen.de/hjaeger/courses/SeminarSpring04/ESNStandardSlides.pdf.
    """
    return mso(n_timesteps=n_timesteps, freqs=[0.2, 0.311], normalize=normalize)


def mso8(n_timesteps: int, normalize: bool = True):
    """Multiple superimposed oscillator task with 8 frequencies [22]_ [24]_

    This is the $MSO$ task with 8 frequencies: :math:`0.2, 0.311, 0.42,
    0.51, 0.63, 0.74, 0.85, 0.97`, as first described in [24]_.

    .. math::

        MSO(t) = \sum_{i=1}^{8} sin(f_i t)

    Parameters
    ----------
    n_timesteps : int
        Number of timesteps to generate
    normalize : bool
        If `True`, scales the range of values in [-1, 1]

    Returns
    -------
    array of shape (n_timesteps, 1)
        Multiple superimposed oscillator timeseries.

    Examples
    --------

    >>> from reservoirpy.datasets import mso8
    >>> timeseries = mso(500)
    >>> print(timeseries.shape)
    (500, 1)

    .. plot::

        from reservoirpy.datasets import mso8
        timeseries = mso8(500)
        plt.figure(figsize=(12, 4))
        plt.plot(timeseries)
        plt.show()

    References
    ----------
    .. [22] Jaeger, H. (2004b). Seminar slides. (Online) Available
            http://www.faculty.iu-bremen.de/hjaeger/courses/SeminarSpring04/ESNStandardSlides.pdf.
    .. [24] Roeschies, B., & Igel, C. (2010). Structure optimization of reservoir networks.
            Logic Journal of IGPL, 18(5), 635-669.
    """
    return mso(
        n_timesteps=n_timesteps,
        freqs=[0.2, 0.311, 0.42, 0.51, 0.63, 0.74, 0.85, 0.97],
        normalize=normalize,
    )



"""
============================================
ReservoirPy Nodes (:mod:`reservoirpy.nodes`)
============================================

.. currentmodule:: reservoirpy.nodes

Reservoirs
==========

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Reservoir - Recurrent pool of leaky integrator neurons
   NVAR - Non-linear Vector Autoregressive machine (NG-RC)
   IPReservoir - Reservoir with intrinsic plasticity learning rule
   LocalPlasticityReservoir - Reservoir with weight plasticity
   ES2N - Edge of Stability Echo State Network

Offline readouts
================

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Ridge - Layer of neurons connected through offline linear regression.
   ScikitLearnNode - Interface for linear models from the scikit-learn library.

Online readouts
===============

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   LMS - Layer of neurons connected through least mean squares learning rule.
   RLS - Layer of neurons connected through recursive least squares learning rule.

Activation functions
====================

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Tanh - Hyperbolic tangent node.
   Sigmoid - Logistic function node.
   Softmax - Softmax node.
   Softplus - Softplus node.
   ReLU - Rectified Linear Unit node.
   Identity - Identity function node.

Input and Output
================

.. autosummary::
   :toctree: generated/
   :template: autosummary/class.rst

   Input - Input node, used to distribute input data to other nodes.
   Output - Output node, used to gather stated from hidden nodes.

"""


# Author: Nathan Trouvain at 16/12/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>

from .activations import Identity, ReLU, Sigmoid, Softmax, Softplus, Tanh
from .es2n import ES2N
from .intrinsic_plasticity import IPReservoir
from .io import Input, Output
from .lif import LIF
from .lms import LMS
from .local_plasticity_reservoir import LocalPlasticityReservoir
from .nvar import NVAR
from .reservoir import Reservoir
from .ridge import Ridge
from .rls import RLS
from .sklearn_node import ScikitLearnNode

__all__ = [
    "Reservoir",
    "Input",
    "Output",
    "Ridge",
    "LMS",
    "RLS",
    "Tanh",
    "Softmax",
    "Softplus",
    "Identity",
    "Sigmoid",
    "ReLU",
    "NVAR",
    "IPReservoir",
    "ScikitLearnNode",
    "LIF",
    "LocalPlasticityReservoir",
    "ES2N",
]



"""
=============================================================
Activations functions (:py:mod:`reservoirpy.activationsfunc`)
=============================================================

Activation functions for reservoir, feedback and output.

.. autosummary::
   :toctree: generated/

    get_function
    identity
    sigmoid
    tanh
    relu
    softmax
    softplus

"""

# Author: Nathan Trouvain at 01/06/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from functools import wraps
from typing import Callable, Union

import numpy as np


def _elementwise(func):
    """Vectorize a function to apply it
    on arrays.
    """
    vect = np.vectorize(func)

    @wraps(func)
    def vect_wrapper(*args, **kwargs):
        u = np.asanyarray(args)
        v = vect(u)
        return v[0]

    return vect_wrapper


def get_function(name: Union[Callable, str]) -> Callable:
    """Return an activation function from name.

    Parameters
    ----------
    name : str, Callable
        Name of the activation function.
        Can be one of {'softmax', 'softplus',
        'sigmoid', 'tanh', 'identity', 'relu'} or
        their respective short names {'smax', 'sp',
        'sig', 'id', 're'}. If `name` is a Callable,
        simply returns `name`.

    Returns
    -------
    callable
        An activation function.
    """
    if isinstance(name, Callable):
        return name

    index = {
        "softmax": softmax,
        "softplus": softplus,
        "sigmoid": sigmoid,
        "tanh": tanh,
        "identity": identity,
        "relu": relu,
        "smax": softmax,
        "sp": softplus,
        "sig": sigmoid,
        "id": identity,
        "re": relu,
    }

    if index.get(name) is None:
        raise ValueError(f"Function name must be one of {[k for k in index.keys()]}")
    else:
        return index[name]


def softmax(x: np.ndarray, beta: float = 1.0) -> np.ndarray:
    """Softmax activation function.

    .. math::

        y_k = \\frac{e^{\\beta x_k}}{\\sum_{i=1}^{n} e^{\\beta x_i}}

    Parameters
    ----------
    x : array
        Input array.
    beta: float, default to 1.0
        Beta parameter of softmax.
    Returns
    -------
    array
        Activated vector.
    """
    _x = np.asarray(x)
    return np.exp(beta * _x) / np.exp(beta * _x).sum(axis=-1, keepdims=True)


@_elementwise
def softplus(x: np.ndarray) -> np.ndarray:
    """Softplus activation function.

    .. math::

        f(x) = \\mathrm{ln}(1 + e^{x})

    Can be used as a smooth version of ReLU.

    Parameters
    ----------
    x : array
        Input array.
    Returns
    -------
    array
        Activated vector.
    """
    return np.log(1 + np.exp(x))


@_elementwise
def sigmoid(x: np.ndarray) -> np.ndarray:
    """Sigmoid activation function.

    .. math::

        f(x) = \\frac{1}{1 + e^{-x}}


    Parameters
    ----------
    x : array
        Input array.
    Returns
    -------
    array
        Activated vector.
    """
    if x < 0:
        u = np.exp(x)
        return u / (u + 1)
    return 1 / (1 + np.exp(-x))


def tanh(x: np.ndarray) -> np.ndarray:
    """Hyperbolic tangent activation function.

    .. math::

        f(x) = \\frac{e^x - e^{-x}}{e^x + e^{-x}}

    Parameters
    ----------
    x : array
        Input array.
    Returns
    -------
    array
        Activated vector.
    """
    return np.tanh(x)


@_elementwise
def identity(x: np.ndarray) -> np.ndarray:
    """Identity function.

    .. math::

        f(x) = x

    Provided for convenience.

    Parameters
    ----------
    x : array
        Input array.
    Returns
    -------
    array
        Activated vector.
    """
    return x


@_elementwise
def relu(x: np.ndarray) -> np.ndarray:
    """ReLU activation function.

    .. math::

        f(x) = x ~~ \\mathrm{if} ~~ x > 0 ~~ \\mathrm{else} ~~ 0

    Parameters
    ----------
    x : array
        Input array.
    Returns
    -------
    array
        Activated vector.
    """
    if x < 0:
        return 0.0
    return x



# Author: Nathan Trouvain at 25/03/2022 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_equal

from reservoirpy.utils.random import noise, rand_generator, set_seed


def test_set_seed():
    set_seed(45)
    from reservoirpy.utils.random import __SEED

    assert __SEED == 45

    with pytest.raises(TypeError):
        set_seed("foo")


def test_random_generator_from_seed():

    gen1 = rand_generator(123)
    gen2 = np.random.default_rng(123)

    assert gen1.integers(1000) == gen2.integers(1000)


def test_noise():

    rng = np.random.default_rng(123)

    a = noise(rng, gain=0.0)
    assert_equal(a, np.zeros((1,)))

    rng = np.random.default_rng(123)

    a = noise(rng, dist="uniform", gain=2.0)
    b = 2.0 * np.random.default_rng(123).uniform()

    assert_equal(a, b)

    a = noise(rng, dist="uniform", gain=2.0)
    b = noise(rng, dist="uniform", gain=2.0)

    assert np.any(np.abs(a - b) > 1e-8)



# Author: Nathan Trouvain at 24/09/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
# Author: Nathan Trouvain at 06/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from joblib import Parallel, delayed
from numpy.testing import assert_array_equal

from reservoirpy.nodes import Ridge


def test_ridge_init():
    x: np.ndarray = np.ones((100,))
    X: np.ndarray = np.ones((120, 100))
    Y: np.ndarray = np.ones((120, 10))

    node = Ridge(ridge=1e-8)
    node.initialize(X, Y)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8

    node = Ridge(ridge=1e-8, output_dim=10)
    node.initialize(X)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8

    with pytest.raises(ValueError):
        node = Ridge(ridge=1e-8, input_dim=100, output_dim=10)
        node(x)

    node = Ridge(ridge=1e-8, Wout=np.ones((100, 10)), bias=np.ones((10,)))
    node.run(X)
    assert node.output_dim == 10
    assert node.input_dim == 100
    assert node.ridge == 1e-8


def test_ridge_worker():
    X: np.ndarray = 2 * np.ones((120, 100))
    Y: np.ndarray = 2 * np.ones((120, 10))

    node = Ridge(1e-6)

    XXT, YXT, x_sum, y_sum, sample_size = node.worker(X, Y)
    assert XXT.shape == (100, 100)
    assert YXT.shape == (100, 10)
    assert_array_equal(x_sum, 2 * 120 * np.ones((100,)))
    assert_array_equal(y_sum, 2 * 120 * np.ones((10,)))
    assert_array_equal(sample_size, 120)


def test_ridge_fit():
    X: np.ndarray = np.ones((120, 100))
    Y: np.ndarray = np.ones((120, 10))

    node = Ridge(1e-6)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert np.any(node.bias != np.zeros((10,)))

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, np.zeros((10,)))


def test_ridge_fit_multiseries():
    X: np.ndarray = np.ones((15, 12, 100))
    Y: np.ndarray = np.ones((15, 12, 10))

    node = Ridge(1e-6)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert np.any(node.bias != np.zeros((10,)))

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, np.zeros((10,)))


def test_ridge_fit_parallel():
    X: np.ndarray = np.ones((15, 12, 100))
    Y: np.ndarray = np.ones((15, 12, 10))

    node = Ridge(1e-6)
    node.fit(X, Y, workers=-1)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert node.bias.shape == (10,)
    assert np.any(node.bias != np.zeros((10,)))

    node = Ridge(1e-6, fit_bias=False)
    node.fit(X, Y)
    assert node.input_dim == 100
    assert node.output_dim == 10
    assert node.Wout.shape == (100, 10)
    assert_array_equal(node.bias, np.zeros((10,)))


def test_parallel():
    process_count = 16

    rng = np.random.default_rng(seed=42)
    x = rng.random((40000, 10))
    y = x[:, 2::-1] + rng.random((40000, 3)) / 10
    x_run = rng.random((20, 10))

    def run_ridge(i):
        readout = Ridge(ridge=1e-8)
        return readout.fit(x, y).run(x_run)

    parallel = Parallel(n_jobs=process_count, return_as="generator")
    results = list(parallel(delayed(run_ridge)(i) for i in range(process_count)))

    for result in results:
        assert np.all(result == results[0])



from typing import Mapping, Sequence

from reservoirpy.type import is_array


def check_timestep(x, *, expected_dim=None):
    if not is_array(x):
        raise TypeError(f"Input but be an array, got a {type(x)}.")
    if not x.ndim == 1:
        raise ValueError(
            f"Input but be a 1-dimensional array. Received an array of shape {x.shape}."
        )
    if expected_dim is not None and not x.shape == (expected_dim,):
        raise ValueError(f"Expected input of shape {(expected_dim,)}, got {x.shape}.")


def is_timestep(x):
    try:
        check_timestep(x)
        return True
    except:
        return False


def check_timeseries(x, *, expected_dim=None, expected_length=None):
    if not is_array(x):
        raise TypeError(f"Input but be an array, got a {type(x)}.")
    if not x.ndim == 2:
        raise ValueError(
            f"Input but be a 2-dimensional array. Got array of shape {x.shape}."
        )
    if expected_length is not None and not x.shape[0] == expected_length:
        raise ValueError(
            f"Expected timeseries of length {expected_length}, got {x.shape[0]}."
        )
    if expected_dim is not None and not x.shape[1] == expected_dim:
        raise ValueError(
            f"Expected feature dimension to be {expected_dim}, got {x.shape[1]}."
        )


def is_timeseries(x):
    try:
        check_timeseries(x)
        return True
    except:
        return False


def check_multiseries(x, *, expected_dim=None, expected_length=None):
    if isinstance(x, Sequence):
        for ts in x:
            check_timeseries(
                ts, expected_dim=expected_dim, expected_length=expected_length
            )
            if expected_dim is None:
                expected_dim = ts.shape[1]
    elif is_array(x):
        if not x.ndim == 3:
            raise ValueError(
                f"Input but be a 3-dimensional array. Got array of shape {x.shape}."
            )
        if expected_length is not None and not x.shape[1] == expected_length:
            raise ValueError(
                f"Expected timeseries of length {expected_length}, got {x.shape[1]}."
            )
        if expected_dim is not None and not x.shape[2] == expected_dim:
            raise ValueError(
                f"Expected feature dimension to be {expected_dim}, got {x.shape[2]}."
            )
    else:
        raise TypeError(
            f"Expected a 3-dimensional array or a sequence of array, got {type(x)}."
        )


def is_multiseries(x):
    try:
        check_multiseries(x)
        return True
    except:
        return False


def check_node_input(x, *, expected_dim=None, expected_length=None):
    if isinstance(x, Sequence):
        for ts in x:
            check_timeseries(
                ts, expected_dim=expected_dim, expected_length=expected_length
            )
            if (
                expected_dim is None
            ):  # Ensure all series have the same number of features
                expected_dim = ts.shape[1]
    elif is_array(x):
        if x.ndim == 3:
            check_multiseries(
                x, expected_dim=expected_dim, expected_length=expected_length
            )
        elif x.ndim == 2:
            check_timeseries(
                x, expected_dim=expected_dim, expected_length=expected_length
            )
        else:
            raise ValueError(
                f"Input but be a (2 or 3)-dimensional array. Got array of shape {x.shape}."
            )
    else:
        raise TypeError(f"Expected an array or a sequence of array, got {type(x)}.")


def is_node_input(x):
    try:
        check_node_input(x)
        return True
    except:
        return False


def check_model_timestep(x, *, expected_inputs=None, expected_dim=None):
    if isinstance(x, Mapping):
        if expected_inputs is not None and set(expected_inputs) != set(x.keys()):
            raise ValueError(
                f"Input does not match expected node names: {set(expected_inputs)} != {set(x.keys())}"
            )
        if isinstance(expected_dim, Mapping):
            for name in x:
                check_timestep(x[name], expected_dim=expected_dim[name])
        else:
            for name in x:
                check_timestep(x[name], expected_dim=expected_dim)
    elif is_array(x):
        if isinstance(expected_dim, Mapping):
            check_timestep(x, expected_dim=next(iter(expected_dim.values())))
        else:
            check_timestep(x, expected_dim=expected_dim)
    else:
        raise TypeError(f"Expected an array or a mapping of arrays, but got {type(x)}")


def is_model_timestep(x):
    try:
        check_model_timestep(x)
        return True
    except:
        return False


def check_model_input(x, *, expected_dim=None, expected_length=None):
    if isinstance(x, Mapping):
        for name in x:
            node_dim = expected_dim[name] if expected_dim is not None else None
            check_node_input(
                x[name], expected_dim=node_dim, expected_length=expected_length
            )
    elif is_array(x) or isinstance(x, Sequence):
        if isinstance(expected_dim, Mapping):
            if len(expected_dim) == 1:
                expected_dim = next(iter(expected_dim.values()))
            else:
                raise TypeError(
                    f"Expected a mapping of node inputs ({list(expected_dim.keys())}), but got {type(x)}."
                )
        check_node_input(x, expected_dim=expected_dim, expected_length=expected_length)
    else:
        raise TypeError(f"Expected an array or a mapping of arrays, but got {type(x)}")


def is_model_input(x):
    try:
        check_model_input(x)
        return True
    except:
        return False



# Author: Nathan Trouvain at 18/11/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
from math import comb

import numpy as np

from reservoirpy.nodes import NVAR


def _get_output_dim(input_dim, delay, order):
    linear_dim = delay * input_dim
    nonlinear_dim = comb(linear_dim + order - 1, order)
    return int(linear_dim + nonlinear_dim)


def test_nvar():
    node = NVAR(3, 2)

    data = np.ones((10,))
    res = node(data)

    assert node.store is not None
    assert node.strides == 1
    assert node.delay == 3
    assert node.order == 2
    assert node.input_dim == 10
    assert node.output_dim == _get_output_dim(10, 3, 2)
    assert res.shape == (_get_output_dim(10, 3, 2),)

    data = np.ones((1000, 10))
    res = node.run(data)

    assert res.shape == (1000, _get_output_dim(10, 3, 2))



import importlib.resources
import sys

import numpy as np

FILENAME = "santafe_laser.npy"


def santafe_laser():
    """Load the Santa-Fe laser [25]_ dataset

    The Santa-Fe laser dataset is a recording of a far-infrared laser activity
    in a chaotic state.

    This is a static dataset of 10,093 timesteps and it is commonly used as a
    forecasting task in reservoir computing benchmarks.

    This specific dataset was taken from James McNames laboratory website [1]_.

    .. plot::

        from reservoirpy.datasets import santafe_laser
        plt.figure(figsize=(8, 4))
        plt.plot(santafe_laser())
        plt.xlabel("Time (a.u.)"); plt.xticks([])
        plt.show()

    Returns
    -------
    X
        Array of shape (10_093, 1).

    References
    ----------
    .. [25] Weigend, A. S. (2018). Time series prediction: forecasting the
            future and understanding the past. Routledge.
    .. [1]  McNames, J. (n.d.). McNames Data Sets.
            http://web.cecs.pdx.edu/mcnames/DataSets/index.html

    """
    binary_data = importlib.resources.files(__package__).joinpath(FILENAME).open("rb")

    numpy_data = np.load(binary_data)

    return numpy_data



import numpy as np
import pytest
from scipy.sparse import csr_array

from ..nodes import Reservoir, Ridge
from ..observables import (
    effective_spectral_radius,
    memory_capacity,
    mse,
    nrmse,
    rmse,
    rsquare,
    spectral_radius,
)


@pytest.mark.parametrize(
    "obs,ytest,ypred,kwargs,expects",
    [
        (mse, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5], {}, None),
        (rmse, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5], {}, None),
        (nrmse, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5], {}, None),
        (nrmse, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5], {"norm": "var"}, None),
        (nrmse, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5], {"norm": "q1q3"}, None),
        (nrmse, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5], {"norm": "foo"}, "raise"),
        (nrmse, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5], {"norm_value": 3.0}, None),
        (rsquare, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5], {}, None),
        (mse, [1.0, 2.0, 3.0, 4.0], [1.5, 2.5, 3.5], {}, "raise"),
        (rmse, [[1.0, 2.0, 3.0]], [1.5, 2.5, 3.5], {}, "raise"),
        (nrmse, [1.0, 2.0, 3.0], [1.5, 2.5, 3.5, 4.2], {}, "raise"),
        (rsquare, [1.0, 2.0, 3.0, 0.0], [1.5, 2.5, 3.5], {}, "raise"),
    ],
)
def test_observable(obs, ytest, ypred, kwargs, expects):
    if expects == "raise":
        with pytest.raises(ValueError):
            obs(ytest, ypred, **kwargs)
    else:
        m = obs(ytest, ypred, **kwargs)
        assert isinstance(m, float)


def test_spectral_radius():
    rng = np.random.default_rng(1234)

    w = rng.uniform(size=(100, 100))

    rho = spectral_radius(w)

    assert isinstance(rho, float)

    idxs = rng.random(size=(100, 100))
    w[idxs < 0.5] = 0
    w = csr_array(w)

    rho = spectral_radius(w)

    assert isinstance(rho, float)

    rho = spectral_radius(w, maxiter=500)

    assert isinstance(rho, float)

    with pytest.raises(ValueError):  # not a square matrix
        w = rng.uniform(size=(5, 100))
        rho = spectral_radius(w)


@pytest.mark.parametrize("observable", [mse, rmse, nrmse, rsquare])
def test_dimensionwise(observable):
    rng = np.random.default_rng(1234)
    # single series
    y1 = rng.uniform(size=(100, 2))
    noise = rng.uniform(size=(100, 2))
    y2 = y1 + noise

    total_score = observable(y_true=y1, y_pred=y2)
    dimensionwise_score = observable(y_true=y1, y_pred=y2, dimensionwise=True)

    assert isinstance(total_score, float)
    assert isinstance(dimensionwise_score, np.ndarray)
    assert dimensionwise_score.shape == (2,)

    # multi-series
    y1 = rng.uniform(size=(3, 100, 2))
    noise = rng.uniform(size=(3, 100, 2))
    y2 = y1 + noise

    total_score = observable(y_true=y1, y_pred=y2)
    dimensionwise_score = observable(y_true=y1, y_pred=y2, dimensionwise=True)

    assert isinstance(total_score, float)
    assert isinstance(dimensionwise_score, np.ndarray)
    assert dimensionwise_score.shape == (2,)


def test_memory_capacity():
    N = 100
    k_max = 2 * N
    model = Reservoir(N, seed=1) >> Ridge(ridge=1e-4)
    mc = memory_capacity(model, k_max=k_max, seed=1)
    mcs = memory_capacity(model, k_max=k_max, as_list=True, seed=1)

    assert isinstance(mc, float)
    assert 0 < mc < k_max
    assert isinstance(mcs, np.ndarray)
    assert mcs.shape == (k_max,)
    assert np.abs(mc - np.sum(mcs)) < 1e-10
    for mc_k in mcs:
        assert 0 < mc_k < 1

    _ = memory_capacity(model, k_max=300, test_size=200)

    # longer lag than the series length
    with pytest.raises(ValueError):
        _ = memory_capacity(model, k_max=300, series=np.ones((100, 1)))
    # invalid test_size argument
    with pytest.raises(ValueError):
        _ = memory_capacity(model, k_max=300, test_size=23.41)
    with pytest.raises(ValueError):
        _ = memory_capacity(model, k_max=300, test_size=None)


def test_effective_spectral_radius():
    reservoir = Reservoir(200, sr=1.0, lr=0.3)
    reservoir.initialize(np.ones((1, 1)))

    esr = effective_spectral_radius(W=reservoir.W, lr=reservoir.lr)
    assert isinstance(esr, float)



import os
from contextlib import contextmanager

import numpy as np
import pytest
from joblib import Memory
from numpy.testing import assert_allclose

from reservoirpy import _TEMPDIR, datasets


@contextmanager
def no_cache():
    """Disable caching temporarily when running tests"""
    datasets._chaos.memory = Memory(location=None)
    yield
    datasets._chaos.memory = Memory(os.path.join(_TEMPDIR, "datasets"), verbose=0)


@pytest.mark.parametrize(
    "dataset_func",
    [
        datasets.henon_map,
        datasets.logistic_map,
        datasets.lorenz,
        datasets.mackey_glass,
        datasets.multiscroll,
        datasets.doublescroll,
        datasets.rabinovich_fabrikant,
        datasets.lorenz96,
        datasets.rossler,
        datasets.kuramoto_sivashinsky,
        datasets.mso2,
        datasets.mso8,
    ],
)
def test_generation(dataset_func):
    with no_cache():
        timesteps = 2000
        X = dataset_func(timesteps)

    assert isinstance(X, np.ndarray)
    assert len(X) == timesteps


def test_generation_narma():
    with no_cache():
        timesteps = 2000
        X = datasets.narma(timesteps)

    assert isinstance(X[0], np.ndarray)
    assert isinstance(X[1], np.ndarray)
    assert len(X[1]) == timesteps


@pytest.mark.parametrize(
    "dataset_func,kwargs,expected",
    [
        (datasets.logistic_map, {"r": -1}, ValueError),
        (datasets.logistic_map, {"x0": 1}, ValueError),
        (datasets.mackey_glass, {"history": np.ones((20,))}, np.ndarray),
        (datasets.mackey_glass, {"history": np.ones((10,))}, ValueError),
        (datasets.narma, {"seed": 1234}, tuple),
        (datasets.mackey_glass, {"seed": 1234}, np.ndarray),
        (datasets.mackey_glass, {"seed": None}, np.ndarray),
        (datasets.mackey_glass, {"tau": 0}, np.ndarray),
        (datasets.narma, {"seed": 1234}, tuple),
        (datasets.lorenz96, {"N": 1}, ValueError),
        (datasets.lorenz96, {"x0": [0.1, 0.2, 0.3, 0.4, 0.5], "N": 4}, ValueError),
        (datasets.rossler, {"x0": [0.1, 0.2]}, ValueError),
        (
            datasets.kuramoto_sivashinsky,
            {"x0": np.random.normal(size=129), "N": 128},
            ValueError,
        ),
        (
            datasets.kuramoto_sivashinsky,
            {"x0": np.random.normal(size=128), "N": 128},
            np.ndarray,
        ),
        (datasets.mso, {"freqs": [0.1, 0.2, 0.3]}, np.ndarray),
        (datasets.mso, {"freqs": []}, np.ndarray),
        (datasets.mso2, {"normalize": False}, np.ndarray),
    ],
)
def test_kwargs(dataset_func, kwargs, expected):
    timesteps = 2000

    if not issubclass(expected, Exception):
        X = dataset_func(timesteps, **kwargs)
        assert isinstance(X, expected)
        if expected is np.ndarray:
            assert X.shape[-2] == timesteps
        if expected is tuple:
            assert X[1].shape[-2] == timesteps
    else:
        with pytest.raises(expected):
            dataset_func(timesteps, **kwargs)


@pytest.mark.parametrize("dataset_func", [datasets.mackey_glass])
def test_seed(dataset_func):
    x1 = dataset_func(200)
    x2 = dataset_func(200)

    assert_allclose(x1, x2)


@pytest.mark.parametrize("dataset_func", [datasets.mackey_glass])
def test_reseed(dataset_func):
    s = datasets.get_seed()
    assert s == datasets._seed._DEFAULT_SEED

    x1 = dataset_func(200)

    datasets.set_seed(1234)
    assert datasets._seed._DEFAULT_SEED == 1234
    assert datasets.get_seed() == 1234

    x2 = dataset_func(200)

    assert (np.abs(x1 - x2) > 1e-3).sum() > 0


@pytest.mark.parametrize("dataset_func", [datasets.mackey_glass, datasets.lorenz])
def test_to_forecasting(dataset_func):
    x = dataset_func(200)

    x, y = datasets.to_forecasting(x, forecast=5)

    assert x.shape[0] == 200 - 5
    assert y.shape[0] == 200 - 5
    assert x.shape[0] == y.shape[0]


@pytest.mark.parametrize("dataset_func", [datasets.mackey_glass, datasets.lorenz])
def test_to_forecasting_with_test(dataset_func):
    x = dataset_func(200)

    x, xt, y, yt = datasets.to_forecasting(x, forecast=5, test_size=10)

    assert x.shape[0] == 200 - 5 - 10
    assert y.shape[0] == 200 - 5 - 10
    assert x.shape[0] == y.shape[0]
    assert xt.shape[0] == yt.shape[0] == 10


def test_japanese_vowels():
    X_train, X_test, Y_train, Y_test = datasets.japanese_vowels(reload=True)

    assert len(X_train) == 270 == len(Y_train)
    assert len(X_test) == 370 == len(Y_test)

    assert Y_train[0].shape == (1, 9)

    X_train, X_test, Y_train, Y_test = datasets.japanese_vowels(repeat_targets=True)

    assert Y_train[0].shape == (X_train[0].shape[0], 9)

    X, X_test, Y, Y_test = datasets.japanese_vowels(one_hot_encode=False)

    assert Y[0].shape == (1, 1)


def test_santafe_laser():
    timeseries = datasets.santafe_laser()

    assert timeseries.shape == (10_093, 1)


def test_one_hot_encode():
    classes = ["green", "blue", "black", "white", "purple"]
    n = 82
    m = 113
    rng = np.random.default_rng(seed=1)
    n_classes = len(classes)

    y = rng.choice(classes, size=(n,), replace=True)
    y_encoded, classes = datasets.one_hot_encode(y)
    assert len(classes) == n_classes
    assert isinstance(y_encoded, np.ndarray) and y_encoded.shape == (n, n_classes)

    y = rng.choice(classes, size=(n, 1), replace=True)
    y_encoded, classes = datasets.one_hot_encode(y)
    assert len(classes) == n_classes
    assert isinstance(y_encoded, np.ndarray) and y_encoded.shape == (n, n_classes)

    y = list(rng.choice(classes, size=(n,), replace=True))
    y_encoded, classes = datasets.one_hot_encode(y)
    assert len(classes) == n_classes
    assert isinstance(y_encoded, np.ndarray) and y_encoded.shape == (n, n_classes)

    y = rng.choice(classes, size=(n, m), replace=True)
    y_encoded, classes = datasets.one_hot_encode(y)
    assert len(classes) == n_classes
    assert isinstance(y_encoded, np.ndarray) and y_encoded.shape == (n, m, n_classes)

    y = rng.choice(classes, size=(n, m, 1), replace=True)
    y_encoded, classes = datasets.one_hot_encode(y)
    assert len(classes) == n_classes
    assert isinstance(y_encoded, np.ndarray) and y_encoded.shape == (n, m, n_classes)

    y = list([rng.choice(classes, size=(m + i, 1), replace=True) for i in range(n)])
    y_encoded, classes = datasets.one_hot_encode(y)
    assert len(classes) == n_classes
    assert isinstance(y_encoded, list)
    assert len(y_encoded) == n
    assert y_encoded[-1].shape == (m + n - 1, n_classes)

    y = list([rng.choice(classes, size=(m + i,), replace=True) for i in range(n)])
    y_encoded, classes = datasets.one_hot_encode(y)
    assert len(classes) == n_classes
    assert isinstance(y_encoded, list)
    assert len(y_encoded) == n
    assert y_encoded[-1].shape == (m + n - 1, n_classes)


def test_from_aeon_classification():
    n_timeseries = 10
    n_timesteps = 100
    n_dimensions = 3
    X_aeon = np.zeros((n_timeseries, n_dimensions, n_timesteps))
    X_aeon[0, 1, 2] = np.pi

    X_rpy = datasets.from_aeon_classification(X_aeon)

    assert X_rpy.shape == (n_timeseries, n_timesteps, n_dimensions)
    assert X_rpy[0, 2, 1] == np.pi

    # variable length collections
    X_aeon_list = [np.zeros((n_dimensions, 10 + i)) for i in range(10)]
    X_aeon_list[0][1, 2] = np.pi

    X_rpy_list = datasets.from_aeon_classification(X_aeon_list)

    assert len(X_rpy_list) == len(X_aeon_list)
    assert X_rpy_list[-1].shape == X_aeon_list[-1].shape[::-1]
    assert X_rpy[0][2, 1] == np.pi

    X_aeon_invalid = True
    with pytest.raises(TypeError):
        datasets.from_aeon_classification(X_aeon_invalid)

    X_aeon_array_like = range(10)
    with pytest.raises(ValueError):
        datasets.from_aeon_classification(X_aeon_array_like)



# Author: Nathan Trouvain at 06/08/2021 <nathan.trouvain@inria.fr>
# Licence: MIT License
# Copyright: Xavier Hinaut (2018) <xavier.hinaut@inria.fr>
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from reservoirpy.activationsfunc import relu
from reservoirpy.mat_gen import bernoulli, zeros
from reservoirpy.nodes import Reservoir


def test_reservoir_init():
    node = Reservoir(100, lr=0.8, bias=zeros)

    data = np.ones((10,))
    res = node(data)

    assert node.W.shape == (100, 100)
    assert node.Win.shape == (100, 10)
    assert node.lr == 0.8
    assert node.units == 100

    data = np.ones((10000, 10))
    res = node.run(data)

    assert res.shape == (10000, 100)

    with pytest.raises(ValueError):
        Reservoir()

    res = Reservoir(100, activation="relu")
    assert id(res.activation) == id(relu)


def test_reservoir_init_from_lr_is_arrays():
    lr = np.ones((100,)) * 0.5
    input_scaling = np.ones((10,)) * 0.8
    node = Reservoir(100, lr=lr, input_scaling=input_scaling)

    data = np.ones((2, 10))
    res = node.run(data)

    assert node.W.shape == (100, 100)
    assert node.Win.shape == (100, 10)
    assert_array_equal(node.lr, np.ones(100) * 0.5)
    assert_array_equal(node.input_scaling, np.ones(10) * 0.8)


def test_reservoir_init_from_matrices():
    Win = np.ones((100, 10))

    node = Reservoir(100, lr=0.8, Win=Win, bias=bernoulli)

    data = np.ones((10,))
    res = node(data)

    assert node.W.shape == (100, 100)
    assert_array_equal(node.Win, Win)
    assert node.lr == 0.8
    assert node.units == 100

    data = np.ones((10000, 10))
    res = node.run(data)

    assert res.shape == (10000, 100)

    Win = np.ones((100, 10))
    bias = np.ones((100,))

    node = Reservoir(100, lr=0.8, Win=Win, bias=bias)

    data = np.ones((10,))
    res = node(data)

    assert node.W.shape == (100, 100)
    assert node.lr == 0.8
    assert node.units == 100
    assert node.output_dim == 100

    data = np.ones((10000, 10))
    res = node.run(data)

    assert res.shape == (10000, 100)

    with pytest.raises(ValueError):  # Shape override (matrix.shape > units parameter)
        data = np.ones((10))
        W = np.ones((10, 10))
        res = Reservoir(100, W=W)
        _ = res(data)

    with pytest.raises(ValueError):  # Bad matrix shape
        W = np.ones((10, 11))
        res = Reservoir(W=W)
        res(data)

    with pytest.raises(ValueError):  # Bad matrix format
        res = Reservoir(100, W=1.0)
        res(data)

    with pytest.raises(ValueError):  # Bad Win shape
        res = Reservoir(100, Win=np.ones((100, 20)))
        res(data)

    with pytest.raises(ValueError):  # Bad Win shape
        res = Reservoir(100, Win=np.ones((101, 10)))
        res(data)

    with pytest.raises(ValueError):  # Bad matrix format
        res = Reservoir(100, Win=1.0)
        res(data)


def test_reservoir_bias():
    node = Reservoir(100, lr=0.8, bias=zeros)
    data = np.ones((10,))
    res = node(data)
    assert node.W.shape == (100, 100)
    assert node.Win.shape == (100, 10)
    assert node.bias.shape == (100,)
    assert_array_equal(node.bias, np.zeros((100,)))
    assert node.lr == 0.8
    assert node.units == 100

    node = Reservoir(100, lr=0.8, bias=bernoulli)
    data = np.ones((10,))
    res = node(data)
    assert node.bias.shape == (100,)

    bias = np.ones((100,))
    node = Reservoir(100, bias=bias)
    res = node(data)
    assert_array_equal(node.bias, bias)

    bias = np.ones((100,))
    node = Reservoir(100, bias=bias)
    res = node(data)
    assert_array_equal(node.bias, bias)

    data = np.zeros((100,))
    node = Reservoir(100, bias=1.0)
    res = node(data)
    assert_array_equal(res, np.tanh(np.ones((100,))))

    with pytest.raises(AssertionError):
        bias = np.ones((100, 1))
        node = Reservoir(100, bias=bias)
        res = node(data)
        assert res.shape == (100,)

    with pytest.raises(ValueError):
        bias = np.ones((100, 2))
        node = Reservoir(100, bias=bias)
        node(data)


def test_reservoir_run():
    x = np.ones((10, 5))

    res = Reservoir(100)
    out = res.run(x)
    assert out.shape == (10, 100)


# def test_reservoir_chain():
#     node1 = Reservoir(100, lr=0.8, input_bias=False)
#     node2 = Reservoir(50, lr=1.0, input_bias=False)

#     data = np.ones((1, 10))
#     res = (node1 >> node2)(data)

#     assert node1.W.shape == (100, 100)
#     assert node1.Win.shape == (100, 10)
#     assert node2.W.shape == (50, 50)
#     assert node2.Win.shape == (50, 100)

#     assert res.shape == (1, 50)


def test_reservoir_seed():
    node1 = Reservoir(100, seed=123)
    node2 = Reservoir(100, seed=123)

    data = np.ones((10, 10))

    assert_array_equal(node1.run(data), node2.run(data))

    node1 = Reservoir(100, seed=123)
    node2 = Reservoir(100, seed=123)

    data = np.ones((10, 10))

    assert_array_equal(node1.run(data), node2.run(data))



from typing import Callable, Iterable, Optional, Sequence, Tuple, Union

import numpy as np
from scipy import linalg

from ..node import ParallelNode
from ..type import NodeInput, State, Timeseries, Timestep, Weights, is_array


class Ridge(ParallelNode):
    """A single layer of neurons learning with Tikhonov linear regression.

    Output weights of the layer are computed following:

    .. math::

        \\hat{\\mathbf{W}}_{out} = \\mathbf{YX}^\\top ~ (\\mathbf{XX}^\\top +
        \\lambda\\mathbf{Id})^{-1}

    Outputs :math:`\\mathbf{y}` of the node are the result of:

    .. math::

        \\mathbf{y} = \\mathbf{W}_{out}^\\top \\mathbf{x} + \\mathbf{b}

    where:
        - :math:`\\mathbf{X}` is the accumulation of all inputs during training;
        - :math:`\\mathbf{Y}` is the accumulation of all targets during training;
        - :math:`\\mathbf{b}` is the first row of :math:`\\hat{\\mathbf{W}}_{out}`;
        - :math:`\\mathbf{W}_{out}` is the rest of :math:`\\hat{\\mathbf{W}}_{out}`.

    If ``fit_bias`` is True, then :math:`\\mathbf{b}` is non-zero, and a constant
    term is added to :math:`\\mathbf{X}` to compute it.


    Parameters
    ----------
    ridge: float, default to 0.0
        L2 regularization parameter.
    fit_bias : bool, default to True
        If True, then a bias parameter will be learned along with output weights.
    Wout : callable or array-like of shape (input_dim, units), optional
        Output weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    bias : callable or array-like of shape (units,), optional
        Bias weights vector or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    output_dim : int, optional
        Number of units in the readout, can be inferred at first call.
    name : str, optional
        Node name.

    Example
    -------

    >>> x = np.random.normal(size=(100, 3))
    >>> noise = np.random.normal(scale=0.1, size=(100, 1))
    >>> y = x @ np.array([[10], [-0.2], [7.]]) + noise + 12.
    >>>
    >>> from reservoirpy.nodes import Ridge
    >>> ridge_regressor = Ridge(ridge=0.001)
    >>>
    >>> ridge_regressor.fit(x, y)
    >>> ridge_regressor.Wout, ridge_regressor.bias
    array([[ 9.992, -0.205,  6.989]]).T, array([[12.011]])
    """

    #: Regularization parameter (:math:`\\lambda`) (0.0 by default).
    ridge: float
    #: If True, learn a bias term (True by default).
    fit_bias: bool
    #: Learned output weights (:math:`\\mathbf{W}_{out}`).
    Wout: Weights
    #: Learned bias (:math:`\\mathbf{b}`).
    bias: Weights

    def __init__(
        self,
        ridge: float = 0.0,
        fit_bias: bool = True,
        Wout: Optional[Union[Weights, Callable]] = None,
        bias: Optional[Union[Weights, Callable]] = None,
        input_dim: Optional[int] = None,
        output_dim: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self.ridge = ridge
        self.fit_bias = fit_bias
        self.Wout = Wout
        self.bias = bias
        self.name = name
        self.initialized = False
        self.state = {}

        # set input_dim/output_dim (if possible)
        self.input_dim = input_dim
        self.output_dim = output_dim
        if is_array(Wout):
            if input_dim is not None and Wout.shape[0] != input_dim:
                raise ValueError(
                    f"Both 'input_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {input_dim} != {Wout.shape[0]}."
                )
            self.input_dim = Wout.shape[0]
            if output_dim is not None and Wout.shape[1] != output_dim:
                raise ValueError(
                    f"Both 'output_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {output_dim} != {Wout.shape[1]}."
                )
            self.output_dim = Wout.shape[1]
        if is_array(bias):
            if output_dim is not None and bias.shape[0] != output_dim:
                raise ValueError(
                    f"Both 'output_dim' and 'Wout' are set but their dimensions doesn't "
                    f"match: {output_dim} != {bias.shape[0]}."
                )
            self.output_dim = bias.shape[0]

    def initialize(
        self,
        x: Union[NodeInput, Timestep],
        y: Optional[Union[NodeInput, Timestep]] = None,
    ):
        self._set_input_dim(x)
        self._set_output_dim(y)

        # initialize matrices
        if isinstance(self.Wout, Callable):
            self.Wout = self.Wout(self.input_dim, self.output_dim)
        if isinstance(self.bias, Callable):
            self.bias = self.bias(self.output_dim)

        self.initialized = True

    def _step(self, state: State, x: Timestep) -> State:
        return {"out": x @ self.Wout + self.bias}

    def _run(self, state: State, x: Timeseries) -> Tuple[State, Timeseries]:
        out = x @ self.Wout + self.bias
        return {"out": out[-1]}, out

    def worker(self, x: Timeseries, y: Timeseries):
        x_sum = np.sum(x, axis=0)
        y_sum = np.sum(y, axis=0)
        sample_size = x.shape[0]
        XXT = x.T @ x
        YXT = x.T @ y
        return XXT, YXT, x_sum, y_sum, sample_size

    def master(self, generator: Iterable):
        XXT = np.zeros((self.input_dim, self.input_dim))
        YXT = np.zeros((self.input_dim, self.output_dim))
        X_sum = 0.0
        Y_sum = 0.0
        total_samples = 0
        ridge_In = self.ridge * np.eye(self.input_dim)

        for (xxt, yxt, x_sum, y_sum, sample_size) in generator:
            XXT += xxt
            YXT += yxt
            X_sum += x_sum
            Y_sum += y_sum
            total_samples = sample_size

        if self.fit_bias:
            X_means = X_sum / total_samples
            Y_means = Y_sum / total_samples
            XXT -= total_samples * np.outer(X_means, X_means)
            YXT -= total_samples * np.outer(X_means, Y_means)

        Wout = linalg.solve(XXT + ridge_In, YXT, assume_a="sym")

        self.Wout = Wout
        if self.fit_bias:
            self.bias = Y_means - X_means @ Wout
        else:
            self.bias = np.zeros((self.output_dim,))



from functools import partial
from typing import Callable, Optional, Sequence, Union

import numpy as np
from numpy.random import Generator

from ..mat_gen import uniform
from ..node import Node
from ..type import NodeInput, State, Timestep, Weights, is_array
from ..utils.random import rand_generator


class LIF(Node):
    """Pool of leaky integrate and fire (LIF) spiking neurons with random recurrent connexions.

    This node is similar to a reservoir (large pool of recurrent, randomly connected neurons),
    but the neurons follows a leaky integrate and fire activity rule.


    Parameters
    ----------
    units : int, optional
        Number of reservoir units. If None, the number of units will be inferred from
        the ``W`` matrix shape.
    inhibitory : float, defaults to 0.0
        Proportion of neurons that have an inhibitory behavior (i.e. negative outgoing
        connections). Must be in :math:`[0, 1]`
    threshold : float, defaults to 1.0
        Limits above which the neurons spikes and returns to zero.
    lr : float or array-like of shape (units,), default to 1.0
        Neurons leak rate. Must be in :math:`[0, 1]`.
    sr : float, defaults to 1.0
        Spectral radius of recurrent weight matrix.
    input_scaling : float or array-like of shape (features,), default to 1.0.
        Input gain. An array of the same dimension as the inputs can be used to
        set up different input scaling for each feature.
    rc_connectivity : float, defaults to 0.1
        Connectivity of recurrent weight matrix, i.e. ratio of reservoir
        neurons connected to other reservoir neurons, including themselves.
        Must be in :math:`]0, 1]`.
    input_connectivity : float, default to 0.1
        Connectivity of input neurons, i.e. ratio of input neurons connected
        to reservoir neurons. Must be in :math:`]0, 1]`.
    Win : callable or array-like of shape (units, features), default to :py:func:`~reservoirpy.mat_gen.uniform` with a
        lower bound of 0.0.
        Input weights matrix or initializer. If a callable (like a function) is used,
        then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    W : callable or array-like of shape (units, units), defaults to :py:func:`~reservoirpy.mat_gen.uniform` with
        a lower bound of 0.0.
        Recurrent weights matrix or initializer. If a callable (like a function) is
        used, then this function should accept any keywords
        parameters and at least two parameters that will be used to define the shape of
        the returned weight matrix.
    input_dim : int, optional
        Input dimension. Can be inferred at first call.
    dtype : Numpy dtype, default to np.float64
        Numerical type for node parameters.
    seed : int or :py:class:`numpy.random.Generator`, optional
        A random state seed, for noise generation.
    name : str, optional
        Node name.

    Note
    ----

    If W, Win, bias or Wfb are initialized with an array-like matrix, then all
    initializers parameters such as spectral radius (``sr``) or input scaling
    (``input_scaling``) are ignored.
    See :py:mod:`~reservoirpy.mat_gen` for more information.

    Example
    -------

    >>> from reservoirpy.nodes import LIF
    >>> liquid = LIF(
    ...     units=100,
    ...     inhibitory=0.1,
    ...     sr=1.0,
    ...     lr=0.2,
    ...     input_scaling=0.5,
    ...     rc_connectivity=1.0,
    ...     input_connectivity=1.0,
    ...     seed=0,
    ... )

    Using the :py:func:`~reservoirpy.datasets.mackey_glass` timeseries:

    >>> from reservoirpy.datasets import mackey_glass
    >>> x = mackey_glass(1000)
    >>> spikes = liquid.run(x)

    .. plot::

        from reservoirpy.nodes import LIF
        liquid = LIF(
            units=100,
            inhibitory=0.1,
            sr=0.5,
            lr=0.2,
            input_scaling=0.5,
            rc_connectivity=1.0,
            input_connectivity=1.0,
        )
        from reservoirpy.datasets import mackey_glass
        x = mackey_glass(1000)
        states = liquid.run(x)
        fig, ax = plt.subplots(6, 1, figsize=(7, 10), sharex=True)
        ax[0].plot(x)
        ax[0].grid()
        ax[0].set_title("Neuron spikes (on Mackey-Glass)")
        for i in range(1, 6):
            ax[i].plot(states[:, i], label=f"Neuron {i}")
            ax[i].legend()
            ax[i].grid()
        ax[-1].set_xlabel("Timesteps")
    """

    #: Number of neuronal units in the reservoir.
    units: int
    #: Proportion of inhibitory neurons. (0.0 by default)
    inhibitory: float
    #: Spike threshold. (1.0 by default)
    threshold: float
    #: Type of matrices elements. By default, ``np.float64``.
    dtype: type
    #: Leaking rate (1.0 by default) (:math:`\mathrm{lr}`).
    lr: Union[float, np.ndarray]
    #: Spectral radius of ``W`` (optional).
    sr: float
    #: Input scaling (float or array) (1.0 by default).
    input_scaling: Union[float, Sequence]
    #: Connectivity (or density) of ``Win`` (0.1 by default).
    input_connectivity: float
    #: Connectivity (or density) of ``Wfb`` (0.1 by default).
    rc_connectivity: float
    #: Input weights matrix (:math:`\mathbf{W}_{in}`).
    Win: Union["Weights", Callable]
    #: Recurrent weights matrix (:math:`\mathbf{W}`).
    W: Union["Weights", Callable]
    #: A random state generator. Used for generating Win and W.
    rng: np.random.Generator

    def __init__(
        self,
        units: Optional[int] = None,
        inhibitory: float = 0.0,
        threshold: float = 1.0,
        lr: float = 0.0,
        sr: float = 1.0,
        input_scaling: Union[float, Sequence] = 1.0,
        rc_connectivity: float = 0.1,
        input_connectivity: float = 0.1,
        Win: Union[Weights, Callable] = partial(uniform, low=0.0),
        W: Union[Weights, Callable] = partial(uniform, low=0.0),
        input_dim: Optional[int] = None,
        dtype: type = np.float64,
        seed: Optional[Union[int, Generator]] = None,
        name: Optional[str] = None,
    ):
        self.inhibitory = inhibitory
        self.threshold = threshold
        self.sr = sr
        self.input_scaling = input_scaling
        self.lr = lr
        self.rc_connectivity = rc_connectivity
        self.input_connectivity = input_connectivity
        self.Win = Win
        self.W = W
        self.dtype = dtype
        self.rng = rand_generator(seed=seed)
        self.name = name
        self.initialized = False

        # set units / output_dim
        if units is None and not is_array(W):
            raise ValueError(
                "'units' parameter must not be None if 'W' parameter is not a matrix."
            )
        if units is not None and is_array(W) and W.shape[-1] != units:
            raise ValueError(
                f"Both 'units' and 'W' are set but their dimensions doesn't match: "
                f"{units} != {W.shape[-1]}."
            )
        self.units = units if units is not None else W.shape[-1]
        self.output_dim = self.units

        # set input_dim (if possible)
        if input_dim is not None and is_array(Win) and Win.shape[-1] != input_dim:
            raise ValueError(
                f"Both 'input_dim' and 'Win' are set but their dimensions doesn't "
                f"match: {input_dim} != {Win.shape[-1]}."
            )
        self.input_dim = Win.shape[-1] if is_array(Win) else input_dim

    def _step(self, state: State, x: Timestep):
        v = state["internal"].copy()
        threshold = self.threshold
        lr = self.lr
        W = self.W
        Win = self.Win
        # leak
        v *= 1 - lr
        # fire
        spikes = (v > threshold).astype(self.dtype)  # threshold
        v[v > threshold] = 0.0
        # integrate
        v += (W @ spikes.T).T
        v += (Win @ x.T).T

        # return spikes
        return {"internal": v, "out": spikes}

    def initialize(
        self,
        x: Optional[Union[NodeInput, Timestep]],
    ):
        self._set_input_dim(x)

        [Win_rng, W_rng, bias_rng] = self.rng.spawn(3)

        if callable(self.Win):
            self.Win = self.Win(
                self.units,
                self.input_dim,
                input_scaling=self.input_scaling,
                connectivity=self.input_connectivity,
                dtype=self.dtype,
                seed=Win_rng,
            )

        if callable(self.W):
            self.W = self.W(
                self.units,
                self.units,
                sr=self.sr,
                connectivity=self.rc_connectivity,
                dtype=self.dtype,
                seed=W_rng,
            )
            n_inhibitory = int(self.inhibitory * self.units)
            self.W[:, :n_inhibitory] *= -1

        self.state = {
            "internal": np.zeros((self.units,)),
            "out": np.zeros((self.units,)),
        }

        self.initialized = True
